import argparse
import logging
import random
import time

from textual import widgets, containers, events, validation, on
from textual.app import App, ComposeResult

from inu import Inu, InuHandler, const, Status
from inu.schema import settings, Heartbeat
from inu.schema.command import Trigger, Jog, Ota, Reboot
from micro_nats import error as mn_error, model
from micro_nats.jetstream.protocol.consumer import Consumer, ConsumerConfig
from micro_nats.util import Time


class InfoWidget(widgets.Static):
    def compose(self) -> ComposeResult:
        # Device status
        stat = widgets.Static(classes="info_sub", id="info_status")
        stat.mount(widgets.Static(f"Status\n ", classes="setting_title"))
        stat.mount(widgets.Checkbox("Enabled", disabled=True, id="stat_enabled"))
        stat.mount(widgets.Checkbox("Locked", disabled=True, id="stat_locked"))
        stat.mount(widgets.Checkbox("Active", disabled=True, id="stat_active"))
        stat.mount(widgets.Static("", id="stat_info"))
        yield stat

        # Device heartbeat
        hb = widgets.Static(classes="info_sub", id="info_hb")
        hb.mount(widgets.Static("Heartbeat\n", classes="setting_title"))
        hb.mount(containers.Horizontal(
            widgets.Static(f" :heart: ", id="hb_heart"),
            widgets.Rule(id="hb_progress", line_style="heavy")
        ))
        hb.mount(widgets.Static("", id="dvc_build"))
        hb.mount(widgets.Static("", id="dvc_address"))
        yield hb

        # Send device commands
        cmd = widgets.Static(classes="info_sub", id="info_cmd")
        cmd.mount(widgets.Static("Command Dispatch", classes="setting_title"))
        cmd.mount(widgets.Button("Toggle Enabled", id="btn_enabled"))
        cmd.mount(widgets.Button("Toggle Locked", id="btn_locked"))
        cmd.mount(widgets.Button("Send Trigger", id="btn_trigger"))
        cmd.mount(widgets.Input("0", validators=validation.Integer(minimum=0, maximum=9999), id="trg_code"))
        cmd.mount(widgets.Button("Interrupt", id="btn_interrupt"))
        yield cmd

        # Send OTA request
        ota = widgets.Static(classes="info_sub", id="info_maintenance")
        ota.mount(widgets.Static("Maintenance", classes="setting_title"))
        ota.mount(widgets.Button("Begin OTA", id="btn_ota"))
        ota.mount(widgets.Input("0", validators=validation.Integer(minimum=0, maximum=99), id="ota_version"))
        ota.mount(widgets.Button("Reboot", id="btn_reboot"))
        yield ota

        # Jog (robotics only)
        jog = widgets.Static(classes="info_sub", id="info_jog")
        jog.mount(widgets.Static("Jog Controls\n", classes="setting_title"))
        jog.mount(widgets.Checkbox("Jog mode", id="btn_jog_mode"))
        jog.mount(widgets.Input("A0", id="jog_device"))
        yield jog


class JogWidget(widgets.Static):
    def compose(self) -> ComposeResult:
        yield widgets.Markdown(f"# JOG MODE ENABLED")
        container = widgets.Static("", classes="info_container")
        container.mount(widgets.Static(f"Press ESC to exit jog mode"))

        jog_grid = widgets.Static(id="jog_grid")
        jog_grid.mount(widgets.Static("Device", classes="setting_subtitle"))
        jog_grid.mount(widgets.Static("", id="jog_device_info", classes="setting_setting"))
        jog_grid.mount(widgets.Static("Distance", classes="setting_subtitle"))
        jog_grid.mount(widgets.Static("1 mm", id="jog_distance", classes="setting_setting"))
        jog_grid.mount(widgets.Static("Speed", classes="setting_subtitle"))
        jog_grid.mount(widgets.Static("2 mm/s", id="jog_speed", classes="setting_setting"))
        container.mount(jog_grid)

        container.mount(widgets.Static(f":arrow_left: and :arrow_right: to decrease/increase jog distance"))
        container.mount(widgets.Static(f":arrow_up: and :arrow_down: to jog the selected device"))

        yield container


class SettingsWidget(widgets.Static):
    def __init__(self, setting_name: str, config: tuple, **kwargs) -> None:
        super().__init__(classes="box", **kwargs)
        (cfg_type, cfg_hint, cfg_min, cfg_max) = config
        self.setting_name = setting_name
        self.setting_type = cfg_type

        self.hint_widget = widgets.Static(f"<{cfg_type}> {cfg_hint}", classes="setting_hint")
        id = f"setting_{setting_name}"

        if cfg_type == "int":
            self.input_widget = widgets.Input(validators=validation.Integer(minimum=cfg_min, maximum=cfg_max), id=id)
        elif cfg_type == "bool":
            self.input_widget = widgets.Checkbox("False", id=id)
        else:
            self.input_widget = widgets.Input(validators=[], id=id)

    def compose(self) -> ComposeResult:
        yield widgets.Static(f"{self.setting_name}", classes="setting_title")
        yield self.hint_widget
        yield self.input_widget

    @on(widgets.Checkbox.Changed)
    def update_checkbox(self, event: widgets.Checkbox.Changed) -> None:
        self.input_widget.label.truncate(0)
        self.input_widget.label.append(str(event.value))

    def set_value(self, val):
        if self.setting_type == 'bool':
            self.input_widget.value = bool(val)
        else:
            self.input_widget.value = str(val)

    def get_value(self):
        if self.setting_type == 'bool':
            return bool(self.input_widget.value)
        elif self.setting_type == 'int':
            # str -> float -> int conversions prevent issues with "12.0" still validating as int
            return int(float(self.input_widget.value))
        else:
            return self.input_widget.value


class Settings(InuHandler, App):
    BINDINGS = [
        ("q", "safe_exit", "exit"),
        ("Q", "exit", "(shift) force exit"),
        ("w", "apply", "write settings"),
    ]
    CSS_PATH = "../../assets/settings.tcss"
    HB_MAX = 23

    # IDs excluded from "making a change"
    EXCLUDED_IDS = ["trg_code", "stat_enabled", "stat_locked", "stat_active"]

    JOG_DIST_STEPS = [1, 5, 10, 25, 100, 250]
    JOG_SPEED_STEPS = [2, 5, 10, 20, 50, 100]

    def __init__(self, args: argparse.Namespace):
        super().__init__()
        self.args = args
        self.logger = logging.getLogger('inu.util.settings')
        self.inu = Inu(const.Context(
            device_id=["settings", f"i{random.randint(1000, 9999)}"],
            nats_server=args.nats,

        ), self)
        self.inu.nats.manager.context.auto_reconnect = False
        self.config = None
        self.config_hint = None
        self.title = "Connecting to NATS.."
        self.device_id = self.args.device_id[0]
        self.record = None
        self.saved = True
        self.jog_mode = False
        self.jog_index = 0

        self.hb_interval = None
        self.last_hb = None

    def compose(self) -> ComposeResult:
        # Header
        header = widgets.Header()
        header.tall = True
        yield header
        yield widgets.Footer()

    async def on_mount(self):
        await self.init()

        await self.mount(widgets.Markdown(self.config_hint, id="config_hint"))
        await self.mount(widgets.Static("", classes="error_hint hidden"))

        container = widgets.Static(id="main_container")
        await container.mount(InfoWidget())

        s_container = widgets.Static(id="settings_widget")
        for config_name, config in self.config.items():
            setting = SettingsWidget(config_name, config)
            setting.set_value(getattr(self.record, config_name))
            await s_container.mount(setting)

        await container.mount(s_container)
        await container.mount(JogWidget(id="jog_widget"))
        await self.mount(container)

        await self.subscribe_info()
        self.set_interval(0.2, self.hb_ticker)

        if self.device_id[0:8] != "robotics":
            self.get_widget_by_id("info_jog").styles.display = "none"

        def set_saved():
            self.saved = True

        self.set_timer(0.01, set_saved)

    async def subscribe_info(self):
        """
        Subscribes to heartbeat & status streams for the device.
        """
        await self.inu.js.consumer.create(
            Consumer(const.Streams.HEARTBEAT, ConsumerConfig(
                filter_subject=const.Subjects.fqs(const.Subjects.HEARTBEAT, self.device_id),
                deliver_policy=ConsumerConfig.DeliverPolicy.NEW,
                ack_wait=Time.sec_to_nano(2),
            )), self.on_hb
        )

        await self.inu.js.consumer.create(
            Consumer(const.Streams.STATUS, ConsumerConfig(
                filter_subject=const.Subjects.fqs(const.Subjects.STATUS, self.device_id),
                deliver_policy=ConsumerConfig.DeliverPolicy.LAST_PER_SUBJECT,
                ack_wait=Time.sec_to_nano(2),
            )), self.on_stat
        )

    async def on_stat(self, msg: model.Message):
        """
        Device status update received.
        """
        await self.inu.js.msg.ack(msg)

        stat = Status(msg.get_payload())
        self.get_widget_by_id("stat_enabled").value = stat.enabled
        self.get_widget_by_id("stat_locked").value = stat.locked
        self.get_widget_by_id("stat_active").value = stat.active
        self.get_widget_by_id("stat_info").update(stat.status)

        self.get_widget_by_id("btn_jog_mode").disabled = stat.enabled

    async def on_hb(self, msg: model.Message):
        """
        Device heartbeat received.
        """
        await self.inu.js.msg.ack(msg)

        hb = Heartbeat(msg.get_payload())
        self.hb_interval = hb.interval
        self.last_hb = time.time()
        self.get_widget_by_id("hb_heart").add_class("beat")
        self.get_widget_by_id("hb_progress").styles.width = 0
        self.get_widget_by_id("dvc_build").update(f"Build: {hb.build}")
        self.get_widget_by_id("dvc_address").update(f"Addr:  {hb.local_addr}")

        def hb_done():
            self.get_widget_by_id("hb_heart").remove_class("beat")

        self.set_timer(0.15, hb_done)

    def hb_ticker(self):
        """
        Updates the HB animation.
        """
        if self.hb_interval is None:
            return

        prog = (time.time() - self.last_hb) / self.hb_interval
        prog_bar = self.get_widget_by_id("hb_progress")
        prog_bar.remove_class("late")
        prog_bar.remove_class("offline")
        if prog <= 1:
            # Device healthy
            prog_bar.styles.width = int(self.HB_MAX * prog)
            prog_bar.styles.margin = 0
        elif prog <= 2:
            # Heartbeat late
            if prog >= 1.25:
                prog_bar.add_class("late")

            prog -= 1
            prog_bar.styles.width = self.HB_MAX - int(self.HB_MAX * prog)
            prog_bar.styles.margin = (0, 0, 0, int(self.HB_MAX * prog))
        elif prog <= 2.:
            # HB very late
            prog_bar.styles.width = 0
            prog_bar.styles.margin = 0
        else:
            # Device considered offline
            prog_bar.styles.width = self.HB_MAX
            prog_bar.styles.margin = 0
            prog_bar.add_class("offline")

    async def _on_key(self, event: events.Key) -> None:
        if self.jog_mode:
            if event.key == "escape":
                self.get_widget_by_id("btn_jog_mode").value = False
                # self.set_jog_mode(False)
            elif event.key == "up":
                await self.do_jog(True)
            elif event.key == "down":
                await self.do_jog(False)
            elif event.key == "left":
                self.jog_index = max(0, self.jog_index - 1)
                self.update_jog_hint()
            elif event.key == "right":
                self.jog_index = min(len(self.JOG_DIST_STEPS) - 1, self.jog_index + 1)
                self.update_jog_hint()

        else:
            if event.key == "escape":
                self.set_focus(None)
            elif event.key == "up":
                self.action_focus_previous()
            elif event.key == "down":
                self.action_focus_next()

    async def do_jog(self, forward: bool):
        distance = self.JOG_DIST_STEPS[self.jog_index] if forward else -self.JOG_DIST_STEPS[self.jog_index]

        jog = Jog(
            distance=distance,
            speed=self.JOG_SPEED_STEPS[self.jog_index],
            device_id=self.get_widget_by_id("jog_device").value,
        )

        await self.inu.nats.publish(
            const.Subjects.fqs([const.Subjects.COMMAND, const.Subjects.COMMAND_JOG], f"central.{self.device_id}"),
            jog.marshal()
        )

    def update_jog_hint(self):
        self.get_widget_by_id("jog_distance").update(f"{self.JOG_DIST_STEPS[self.jog_index]} mm")
        self.get_widget_by_id("jog_speed").update(f"{self.JOG_SPEED_STEPS[self.jog_index]} mm/s")

    @on(widgets.Input.Changed)
    def on_input_changed(self, event: widgets.Input.Changed) -> None:
        if event.input.id in self.EXCLUDED_IDS or event.input.id is None or event.input.id[0:4] == "btn_":
            return

        self.saved = False

    @on(widgets.Checkbox.Changed)
    def on_checkbox_changed(self, event: widgets.Checkbox.Changed) -> None:
        if event.checkbox.id == "btn_jog_mode":
            self.set_jog_mode(event.checkbox.value)

        if event.checkbox.id in self.EXCLUDED_IDS or event.checkbox.id is None or event.checkbox.id[0:4] == "btn_":
            return

        self.saved = False

    def set_jog_mode(self, enabled: bool):
        self.jog_mode = enabled

        main_display = "none" if self.jog_mode else "block"
        jog_display = "block" if self.jog_mode else "none"

        self.get_widget_by_id("jog_device_info").update(self.get_widget_by_id("jog_device").value)

        self.get_widget_by_id("settings_widget").styles.display = main_display
        self.get_widget_by_id("config_hint").styles.display = main_display
        self.get_widget_by_id("info_maintenance").styles.display = main_display
        self.get_widget_by_id("info_cmd").styles.display = main_display
        self.get_widget_by_id("info_jog").styles.display = main_display

        self.get_widget_by_id("jog_widget").styles.display = jog_display

        if self.jog_mode:
            self.set_focus(None)
            self.screen.scroll_to(0, 0, speed=100)

    async def on_button_pressed(self, event: widgets.Button.Pressed) -> None:
        trg = Trigger()
        if event.button.id == "btn_trigger":
            trg.code = int(self.get_widget_by_id("trg_code").value)
        elif event.button.id == "btn_interrupt":
            trg.code = const.TriggerCode.INTERRUPT
        elif event.button.id == "btn_enabled":
            trg.code = const.TriggerCode.ENABLE_TOGGLE
        elif event.button.id == "btn_locked":
            trg.code = const.TriggerCode.LOCK_TOGGLE
        elif event.button.id == "btn_ota":
            # OTA has a special command code, not a trigger
            ota = Ota()
            ota.version = int(self.get_widget_by_id("ota_version").value)
            await self.inu.nats.publish(
                const.Subjects.fqs([const.Subjects.COMMAND, const.Subjects.COMMAND_OTA], f"central.{self.device_id}"),
                ota.marshal()
            )
            return
        elif event.button.id == "btn_reboot":
            # Hard reboot
            reboot = Reboot()
            await self.inu.nats.publish(
                const.Subjects.fqs([const.Subjects.COMMAND, const.Subjects.COMMAND_REBOOT],
                                   f"central.{self.device_id}"),
                reboot.marshal()
            )
            return
        else:
            return

        await self.inu.nats.publish(
            const.Subjects.fqs([const.Subjects.COMMAND, const.Subjects.COMMAND_TRIGGER], f"central.{self.device_id}"),
            trg.marshal()
        )

    async def init(self):
        """
        Execute the bootstrap process.
        """
        if not await self.inu.init():
            self.exit(return_code=1, message="Could not connect to NATS server")

        dvc_id = self.device_id.split(".")

        if len(dvc_id) < 2:
            self.exit(
                return_code=1,
                message="Device IDs must contain at least two namespaces (device_type.device_name)"
            )

        try:
            cls = settings.get_device_settings_class(dvc_id[0])
            try:
                msg = await self.inu.js.msg.get_last(
                    const.Streams.SETTINGS,
                    const.Subjects.fqs(const.Subjects.SETTINGS, self.device_id)
                )
                self.record = cls(msg.get_payload())
            except mn_error.NotFoundError:
                self.record = cls()

            self.config_hint, self.config = settings.get_config_for_device(dvc_id[0])
            self.title = f"I . N . U  -  {self.device_id}"
        except ValueError:
            self.exit(return_code=1, message=f"Unknown device type ({dvc_id[0]}) for provided device ID.")
            return 3

        return 0

    def action_exit(self):
        """
        Immediately exit.
        """
        self.exit(0)

    def action_safe_exit(self):
        """
        Exit only if saved.
        """
        if self.saved:
            self.exit(0)
        else:
            self.set_error_message("not saved; Shift+Q to force exit")

    async def action_save(self):
        """
        Save the record, and quit the application on success.
        """
        if await self.save_record():
            self.exit(0)

    async def action_apply(self):
        """
        Save the record, then do nothing.
        """
        await self.save_record()

    async def save_record(self) -> bool:
        """
        Validate the record content and if valid, save the setting record to NATS.

        Returns True if it was saved, False if it failed validation. Will set the error/success message accordingly.
        """
        err = []
        for node in self.query(SettingsWidget).nodes:
            if hasattr(node.input_widget, 'validate'):
                valid = node.input_widget.validate(node.input_widget.value)
                if valid and not valid.is_valid:
                    err.append(node.setting_name)
                    continue

            setattr(self.record, node.setting_name, node.get_value())

        if len(err):
            errs = "\n * ".join(err)
            self.set_error_message(f"Validation failure on:\n * {errs}")
            return False
        else:
            self.set_working_message("Saving.. " + self.record.marshal())
            await self.inu.nats.publish(
                const.Subjects.fqs(const.Subjects.SETTINGS, self.device_id),
                self.record.marshal()
            )
            await self.inu.log(f"Updated settings for {self.device_id}")
            self.set_success_message("Record saved")
            self.saved = True
            return True

    def set_working_message(self, msg: str):
        """
        Style the error/success hint as 'in progress' and set the message.
        """
        self.clear_error_message()
        w = self.query_one("Static.error_hint")
        w.update(msg)
        w.add_class("working")
        w.remove_class("success")
        w.remove_class("hidden")

    def set_success_message(self, msg: str):
        """
        Style the error/success hint as a 'success' and set the message.
        """
        self.clear_error_message()
        w = self.query_one("Static.error_hint")
        w.update(msg)
        w.add_class("success")
        w.remove_class("working")
        w.remove_class("hidden")

        self.set_timer(3, self.conditional_clear_success_message)

    def set_error_message(self, msg: str):
        """
        Style the error/success hint as an 'error' and set the message.
        """
        self.clear_error_message()
        w = self.query_one("Static.error_hint")
        w.update(msg)
        w.remove_class("success")
        w.remove_class("working")
        w.remove_class("hidden")

    def clear_error_message(self):
        """
        Clears any error/success message hint and hides the element.
        """
        w = self.query_one("Static.error_hint")
        w.update("")
        w.add_class("hidden")
        w.remove_class("success")
        w.remove_class("working")

    def conditional_clear_success_message(self):
        """
        Hides the error/success hint only if it still contains a success message.
        """
        w = self.query_one("Static.error_hint")
        if w.has_class("success"):
            w.update("")
            w.add_class("hidden")



import argparse
import logging
import random

from micro_nats.error import NotFoundError
from micro_nats.jetstream.error import ErrorResponseException
from micro_nats.jetstream.protocol.stream import StreamConfig
from micro_nats.util import Time
from .. import Inu, InuHandler, const
from ..util import Utility


class Bootstrap(Utility, InuHandler):
    def __init__(self, args: argparse.Namespace):
        super().__init__(args)
        self.logger = logging.getLogger('inu.util.bootstrap')
        self.inu = Inu(const.Context(
            device_id=["bootstrap", f"i{random.randint(1000, 9999)}"],
            nats_server=args.nats,
        ), self)

    async def run(self):
        """
        Execute the bootstrap process.
        """
        if not await self.inu.init():
            return

        try:
            if self.args.clean or self.args.delete:
                # Delete any existing streams
                await self.delete_streams()

            if not self.args.delete:
                # Configure all Inu streams
                await self.configure_streams()

            # All done
            await self.inu.log(f"Bootstrap completed ({self.inu.device_id})")

        except ErrorResponseException as e:
            err = e.err_response
            self.logger.error(f"NATS: {err.code}-{err.err_code}: {err.description}")

        except Exception as e:
            self.logger.error(f"Aborting due to errors: {type(e).__name__}: {str(e)}")

    async def delete_streams(self):
        """
        Delete any existing Inu streams.
        """
        self.logger.info("Cleaning streams")

        async def clean(s: str):
            try:
                await self.inu.js.stream.delete(s)
                self.logger.warning(f"stream '{s}' deleted")
            except NotFoundError:
                pass

        await clean(const.Streams.LOGS)
        await clean(const.Streams.ALERTS)
        await clean(const.Streams.STATUS)
        await clean(const.Streams.COMMAND)
        await clean(const.Streams.HEARTBEAT)
        await clean(const.Streams.SETTINGS)

    async def configure_streams(self):
        """
        Created all required Inu NATS streams.
        """
        self.logger.info("Preparing streams")
        # Log stream
        await self.inu.js.stream.create(StreamConfig(
            name=const.Streams.LOGS,
            description="Inu network logs",
            subjects=[const.Subjects.all(const.Subjects.LOG)],
            max_msgs_per_subject=100000,
            retention=StreamConfig.RetentionPolicy.LIMITS,
            storage=StreamConfig.StorageType.FILE,
            discard=StreamConfig.DiscardPolicy.OLD
        ))

        # Alerts stream
        await self.inu.js.stream.create(StreamConfig(
            name=const.Streams.ALERTS,
            description="Inu network alerts",
            subjects=[const.Subjects.all(const.Subjects.ALERT)],
            max_msgs_per_subject=100000,
            retention=StreamConfig.RetentionPolicy.LIMITS,
            storage=StreamConfig.StorageType.FILE,
            discard=StreamConfig.DiscardPolicy.OLD
        ))

        # Status stream
        await self.inu.js.stream.create(StreamConfig(
            name=const.Streams.STATUS,
            description="Inu network device statuses",
            subjects=[const.Subjects.all(const.Subjects.STATUS)],
            max_msgs_per_subject=1000,
            retention=StreamConfig.RetentionPolicy.LIMITS,
            storage=StreamConfig.StorageType.FILE,
            discard=StreamConfig.DiscardPolicy.OLD
        ))

        # Command stream
        await self.inu.js.stream.create(StreamConfig(
            name=const.Streams.COMMAND,
            description="Inu network device commands",
            subjects=[const.Subjects.all(const.Subjects.COMMAND)],
            max_msgs_per_subject=1000,
            retention=StreamConfig.RetentionPolicy.LIMITS,
            storage=StreamConfig.StorageType.FILE,
            discard=StreamConfig.DiscardPolicy.OLD
        ))

        # Heartbeat stream
        await self.inu.js.stream.create(StreamConfig(
            name=const.Streams.HEARTBEAT,
            description="Inu network device heartbeats",
            subjects=[const.Subjects.all(const.Subjects.HEARTBEAT)],
            max_msgs_per_subject=3,
            max_age=Time.sec_to_nano(3 * 24 * 3600),  # 3 days - purge temp clients
            retention=StreamConfig.RetentionPolicy.LIMITS,
            storage=StreamConfig.StorageType.FILE,
            discard=StreamConfig.DiscardPolicy.OLD
        ))

        # Settings stream
        await self.inu.js.stream.create(StreamConfig(
            name=const.Streams.SETTINGS,
            description="Inu network device settings",
            subjects=[const.Subjects.all(const.Subjects.SETTINGS)],
            max_msgs_per_subject=100,
            retention=StreamConfig.RetentionPolicy.LIMITS,
            storage=StreamConfig.StorageType.FILE,
            discard=StreamConfig.DiscardPolicy.OLD
        ))



import asyncio
import binascii
import logging
import os
import struct
import time

import machine
import urequests as requests
from inu import Status
from inu.const import LogLevel


class Updater:
    async def perform_update(self, version: int):
        pass


class OtaUpdater(Updater):
    OTA_VERSION_URL = "https://storage.googleapis.com/inu-ota/{app}/version?v={v}"
    OTA_BUILD_URL = "https://storage.googleapis.com/inu-ota/{app}/build-{version}.ota?v={v}"
    OTA_CHECKSUM_URL = "https://storage.googleapis.com/inu-ota/{app}/build-{version}.checksum?v={v}"

    MAX_RETRIES = 10

    def __init__(self, app):
        from inu.app import InuApp
        self.app: InuApp = app
        self.logger = logging.getLogger("inu.ota")

    async def perform_update(self, version: int):
        """
        Halts operations and downloads the specified OTA update before writing to the system.

        If version is 0, the device will query for the latest version and use that.
        """
        if not self.app.allow_app_tick:
            await self.app.inu.log(f"Ignoring OTA request while device is in maintenance mode")
            return

        # Wait for device to finish whatever its doing first
        if self.app.inu.state.active:
            await self.app.inu.log("OTA update requested, will initiate when idle")
            while self.app.inu.state.active:
                await asyncio.sleep(0.1)

        original_state = Status(
            enabled=self.app.inu.state.enabled,
            active=self.app.inu.state.active,
            status=self.app.inu.state.status
        )
        self.app.allow_app_tick = False

        await self.app.inu.log(f"Applying OTA update for {self.app.inu.app_name} v{version}")
        await self.app.inu.status(enabled=False, active=False, status="Applying OTA update")
        await asyncio.sleep(0.25)  # allow messages to go out

        # NB: INTENTIONALLY NOT MODULARISING to avoid increasing memory footprint
        try:
            # Cache-busting
            v = time.time()

            # Version 0 means use the latest version, grab the latest from GCP -
            if version == 0:
                version = await self.get_latest_version()
                if version <= 0:
                    await self.app.inu.log(f"Error acquiring latest version ({version})", LogLevel.ERROR)
                    await self.abort_update(original_state)
                    return
                await self.app.inu.log(f"Latest version for {self.app.inu.app_name} determined to be {version}")

            # This is prone to a high error rate, set a retry-loop -
            response = None
            err = None
            for i in range(0, self.MAX_RETRIES):
                # Check checksum of OTA package
                self.logger.info("Getting OTA packet checksum..")
                response = requests.get(
                    url=self.OTA_CHECKSUM_URL.format(app=self.app.inu.app_name, version=version, v=v)
                )
                if response.status_code != 200:
                    err = f"Error downloading OTA checksum: {response.status_code}"
                    self.logger.warning(err)
                    continue
                checksum = response.text

                await asyncio.sleep(0)

                # Download OTA package into memory (should be ~ 250kb)
                self.logger.info("Downloading OTA packet..")
                response = requests.get(
                    url=self.OTA_BUILD_URL.format(app=self.app.inu.app_name, version=version, v=v)
                )
                if response.status_code != 200:
                    err = f"Error downloading OTA package: {response.status_code}"
                    self.logger.warning(err)
                    continue

                await asyncio.sleep(0)

                # Validate checksum
                self.logger.info(f"Validating checksum ({len(response.content)} bytes)..")
                ota_checksum = "%08X" % (binascii.crc32(response.content) & 0xFFFFFFFF)
                if ota_checksum != checksum:
                    err = f"OTA checksum mismatch; expected: {checksum}, got: {ota_checksum}"
                    self.logger.warning(err)
                    continue

                err = None
                break

            if isinstance(err, str):
                await self.app.inu.log(err, LogLevel.ERROR)
                await self.abort_update(original_state)
                return

        except Exception as e:
            await self.app.inu.log(f"OTA error - {type(e).__name__}: {e}", LogLevel.ERROR)
            await self.abort_update(original_state)
            return

        await asyncio.sleep(0)

        # BE CAREFUL moving around the response content, don't duplicate the memory profile
        index = 0
        package_version = struct.unpack("<I", response.content[index:index + 4])[0]
        self.logger.info(f"OTA package version: {package_version}")
        if package_version != version:
            await self.app.inu.log(
                f"OTA package version error: expected {version}, got {package_version}", LogLevel.ERROR
            )
            await self.abort_update(original_state)
            return

        index += 4

        def unpack_file():
            nonlocal index, response

            # Unpack filename
            fn_len = struct.unpack("<H", response.content[index:index + 2])[0]
            index += 2
            fn = response.content[index:index + fn_len].decode()
            index += fn_len

            data_len = struct.unpack("<I", response.content[index:index + 4])[0]
            index += 4

            self.logger.info(f"write: {fn} ({data_len} b)")
            self.makedirs(fn)

            with open(fn, "wb") as fp:
                fp.write(response.content[index:index + data_len])

            index += data_len

        try:
            while index < len(response.content):
                unpack_file()
                await asyncio.sleep(0)
        except Exception as e:
            await self.app.inu.log(
                f"Error during OTA application - {type(e).__name__}: {e}; index: {index}",
                LogLevel.FATAL
            )
            await self.abort_update(original_state)
            return

        self.app.inu.state = original_state
        await self.app.inu.status(status="OTA reboot")

        await self.app.inu.log(f"OTA update applied, rebooting")
        await asyncio.sleep(0.5)
        machine.reset()

    @staticmethod
    def makedirs(fn: str):
        """
        Creates all directories for a given filename.
        """
        paths = fn.split("/")
        paths.pop()
        path = ""
        for p in paths:
            path += f"/{p}"
            try:
                os.mkdir(path)
            except OSError:
                # exists
                pass

    async def get_latest_version(self):
        """
        Hit the VERSION URL to determine the latest build for the current app.
        """
        response = requests.get(url=self.OTA_VERSION_URL.format(app=self.app.inu.app_name, v=time.time()))
        if response.status_code != 200:
            return -1

        return int(response.text)

    async def abort_update(self, state: Status):
        """
        Resets the device state after an error early in the OTA process.
        """
        await self.app.inu.log("OTA update aborting, resuming device activity")
        self.app.inu.state = state
        await self.app.inu.status(status="")
        self.app.allow_app_tick = True



from inu import error


class Control:
    """
    Controls represent actions that can be taken and are constructed from a control string. eg:
        SEL A0; MV 800 300; W 2000 INT; MV -800 150 INT
        SEL L0:0: COL 255 0 0; SEL L0:1; COL 0 0 255!
    """
    INTERRUPT_CODE = "INT"
    EXECUTE_CODE = "!"
    DELIMITER = ";"

    def __init__(self, cmd: str):
        # Allow the current operation to be interrupted by an INT signal
        self.allow_int = False
        # Instruct the operation to commit/write/execute - required for light changes, etc.
        self.execute = False
        # The control code
        self.code = None
        # Arguments for the control code
        self.args = []

        if cmd is not None:
            self._parse(cmd)

    def allow_interrupt(self) -> bool:
        """
        Check if this control allowed interruption.
        """
        return self.allow_int

    def _parse(self, cmd: str):
        """
        Breaks down `cmd` into args, separating special operators.
        """
        args = cmd.strip().upper().split(" ")
        if len(args) < 2:
            raise error.Malformed(f"Invalid control string: {cmd}")

        self.code = args.pop(0)

        while len(args):
            arg = args.pop(0)
            if arg == self.INTERRUPT_CODE:
                self.allow_int = True
            elif arg == self.EXECUTE_CODE:
                self.execute = True
            else:
                self.args.append(arg)

        if len(self.args) < 1:
            raise error.Malformed(f"No control arguments: {cmd}")



import asyncio
import logging

from machine import Pin
from ..hardware import MotionSensor


class Pir(MotionSensor):
    def __init__(self, pin: int = 33, pull_mode: int = None):
        self.logger = logging.getLogger('inu.hw.pir')
        self.motion = False
        self.pin = Pin(pin, Pin.IN, pull=pull_mode)

    async def read_loop(self):
        while True:
            self.motion = self.pin.value() == 1
            await asyncio.sleep(0.01)

    def is_motion(self) -> bool:
        return self.motion



import logging

from machine import Pin


class Relay:
    def __init__(self, pin: int = 33, ground: int = None, on_state_change: callable = None):
        self.logger = logging.getLogger('inu.hw.relay')
        self.active = False
        self.state_cb = on_state_change

        self.pin = Pin(pin, Pin.OUT)
        self.pin.off()

        if ground:
            self.ground = Pin(ground, Pin.IN)
        else:
            self.ground = None

    async def on(self):
        self.pin.on()
        self.logger.debug(f"Relay:{self.pin} ACTIVE")

        if self.state_cb:
            await self.state_cb(True)

    async def off(self):
        self.pin.off()
        self.logger.debug(f"Relay:{self.pin} INACTIVE")

        if self.state_cb:
            await self.state_cb(False)

    async def toggle(self):
        if self.active:
            await self.off()
        else:
            await self.on()

    def is_active(self) -> bool:
        return self.active



import time

from machine import I2C
from . import AmbientLightSensor


class BH1750(AmbientLightSensor):
    # BH1750 I2C address
    I2C_ADDR = 0x23

    class CONFIG:
        CONTINUOUS_HIGH_RES_MODE = 0x10

    def __init__(self, i2c_id=1):
        self.i2c = I2C(i2c_id)
        self.config = bytearray(1)
        self.set_config(BH1750.CONFIG.CONTINUOUS_HIGH_RES_MODE)

    def read(self):
        data = self.i2c.readfrom(self.I2C_ADDR, 2)
        return ((data[0] << 8) | data[1]) / 1.2

    def set_config(self, config):
        self.config[0] = config
        self.i2c.writeto(self.I2C_ADDR, self.config)
        time.sleep(0.2)



import json
import logging

from ..error import Malformed

logger = logging.getLogger('inu.schema')


class Schema:
    def __init__(self, bootstrap=None, **kwargs):
        if bootstrap is None:
            pass
        elif isinstance(bootstrap, bytearray):
            self.hydrate(bytes(bootstrap))
        elif isinstance(bootstrap, str) or isinstance(bootstrap, bytes):
            self.hydrate(bootstrap)
        elif isinstance(bootstrap, dict):
            self.hydrate_dict(bootstrap)
        else:
            raise Malformed(f"Invalid type for schema bootstrap: {type(bootstrap)}: {bootstrap}")

        if len(kwargs) > 0:
            self.hydrate_dict(kwargs)

    def hydrate(self, payload: str | bytes):
        """
        Hydrate object from a JSON string.
        """
        d = json.loads(payload)

        if not isinstance(d, dict):
            # Super weird bug whereby the first invocation of json.loads returns a str instead of dict.
            # Running the result through the same call again seems to resolve it. Bug was triggered when payload was
            # a bytes object.
            d = json.loads(d)

        self.hydrate_dict(d)

    def hydrate_dict(self, d: dict):
        """
        Hydrate object from a dictionary.
        """
        if not isinstance(d, dict):
            raise Exception(f"d is not a dict: {type(d).__name__}: {d}")

        for k, v in d.items():
            if k[0] == '_':
                logger.warning(f"Attempted to hydrate illegal property '{k}' on {self.__class__.__name__}")
                continue

            if hasattr(self, k):
                if hasattr(self, f"_set_{k}"):
                    fn = getattr(self, f"_set_{k}")
                    fn(v)
                else:
                    setattr(self, k, v)
            else:
                logger.warning(f"No property '{k}' to hydrate on {self.__class__.__name__}")

        if hasattr(self, "_validate"):
            self._validate()

    def marshal(self) -> str:
        """
        Generate JSON from the object properties.
        """
        d = {}
        for k in dir(self):
            if k[0] == '_':
                continue
            v = getattr(self, k)
            if callable(v):
                continue
            d[k] = v
        return json.dumps(d)


class Heartbeat(Schema):
    interval: int = None
    build: int = None
    local_addr: str = None


class Log(Schema):
    level: str = None
    message: str = None

    def _validate(self):
        if self.level is None:
            raise Malformed("Log level cannot be None")

        if self.message is None:
            raise Malformed("Log message cannot be None")

    def __repr__(self):
        return f"<log level={self.level}; message={self.message}>"


class Alert(Schema):
    priority: int = None
    message: str = None

    def _validate(self):
        if self.priority is None:
            raise Malformed("Priority cannot be None")

        if self.message is None:
            raise Malformed("Alert message cannot be None")



from . import Control
from inu import error

from ..colour import ColourCode


class Colour(Control):
    """
    Set device colour & brightness.

    COL <colour <!>

    colour - #RRGGBB[XX] or R,G,B[,X]
    ! - required to commit the change
    """
    CONTROL_CODE = "COL"
    ALIASES = ["COL", "COLOUR", "COLOR"]

    def __init__(self, ctrl: str = None):
        super().__init__(ctrl)

        if self.code not in self.ALIASES or 1 > len(self.args) > 2:
            raise error.Malformed(f"Invalid {self.CONTROL_CODE} control: {ctrl}")

        self.colour = ColourCode(self.args[0])

    def __repr__(self):
        return f"COL {self.colour}"


class Fx(Control):
    """
    LED FX control.

    FX <colour> <duration> <fx>

    colour - #RRGGBB[XX] or R,G,B[,X]
    duration - time in milliseconds
    FX - one of: FADE, SLIDEL, SLIDER, PULSEL, PULSER
    """
    CONTROL_CODE = "FX"

    class FX:
        FADE = "FADE"
        SLIDE_L = "SLIDEL"
        SLIDE_R = "SLIDER"
        PULSE_L = "PULSEL"
        PULSE_R = "PULSER"

    def __init__(self, ctrl: str = None):
        super().__init__(ctrl)

        if len(self.args) != 3:
            raise error.Malformed(f"Invalid {self.CONTROL_CODE} control: {ctrl}")

        self.colour = ColourCode(self.args[0])

    def get_duration(self) -> int:
        """
        Get the duration value in milliseconds.
        """
        return int(self.args[1])

    def get_fx(self) -> str:
        """
        Get the FX value.
        """
        fx = self.args[2].upper().strip()
        if fx not in [self.FX.FADE, self.FX.SLIDE_L, self.FX.SLIDE_R, self.FX.PULSE_L, self.FX.PULSE_R]:
            return self.FX.FADE
        else:
            return fx

    def __repr__(self):
        return f"FX {self.colour} -> {self.get_fx()}"



import asyncio
import logging
import time

import machine
from ..hardware import RangingSensor


class Sonar(RangingSensor):

    def __init__(self, uart: int = 1, tx: int = None, rx: int = None, baud: int = 9600):
        self.logger = logging.getLogger('inu.hw.sonar')
        self.uart = machine.UART(uart, baudrate=baud, tx=tx, rx=rx)

        self.distance: int | None = None
        self.last_measured: float | None = None

    def get_distance(self) -> int | None:
        """
        Returns the distance in mm.

        Returns None if no measurement has been made.
        """
        return self.distance

    def get_age(self) -> float | None:
        """
        Returns the number of seconds since the last valid measurement.

        Returns None if no measurement has been made.
        """
        if self.last_measured is None:
            return None

        return time.time() - self.last_measured

    @staticmethod
    def is_checksum_valid(data):
        cs = (data[0] + data[1] + data[2]) & 0x00ff
        return cs == data[3]

    async def read_loop(self):
        self.logger.debug("Starting read loop..")
        while True:
            if self.uart.any() >= 128:
                self.logger.warning("UART buffer overflow, flushing")
                await self.flush_uart()

            while self.uart.any() >= 4:
                data = self.uart.read(4)

                if data[0] != 0xff:
                    self.logger.warning(f"UART stream corrupted ({data[0]}), flushing")
                    await self.flush_uart()
                    continue

                if not self.is_checksum_valid(data):
                    self.logger.warning("Checksum error, flushing")
                    await self.flush_uart()
                    continue

                distance = data[1] * 256 + data[2]
                if distance > 0:
                    # a distance of exactly zero is a bad sensor reading
                    self.distance = distance
                    self.last_measured = time.time()

            await asyncio.sleep(0.01)

    async def flush_uart(self):
        buffer_size = self.uart.any()
        if buffer_size:
            self.uart.read(buffer_size)
            self.logger.warning(f"Flushed {buffer_size} bytes")



import argparse
import asyncio
import logging
import random

from inu import Inu, InuHandler, tz
from inu import error, const
from inu.const import Context
from inu.error import BadRequest
from inu.schema import Alert, Log, Heartbeat
from inu.schema.settings import Settings
from inu.util import Utility
from micro_nats import error as mn_error, model
from micro_nats.jetstream.error import ErrorResponseException
from micro_nats.jetstream.protocol.consumer import Consumer, ConsumerConfig
from micro_nats.util import Time


class Monitor(Utility, InuHandler):
    # Number of logged messages (only counts if there is a limit)
    logged: int = 0

    # Max number of logged messages before exiting
    limit: int = 0

    # Exit request (from limit hit)
    exit: bool = False

    def __init__(self, args: argparse.Namespace):
        super().__init__(args)
        self.logger = logging.getLogger('inu.util.monitor')
        self.inu = Inu(Context(
            device_id=["monitor", f"i{random.randint(1000, 9999)}"],
            nats_server=args.nats,
            has_heartbeat=False,
            settings_class=Settings,
        ), self)

        self.limit = int(self.args.limit or 0)

    async def run(self):
        """
        Main monitor daemon.

        Never returns.
        """
        if not await self.inu.init():
            return

        while not self.exit:
            await asyncio.sleep(0.1)

    async def on_settings_updated(self):
        self.logger.debug("New settings received")

    async def on_connect(self, server: model.ServerInfo):
        self.logger.info("Connected to NATS server")

        stream_tally = 0
        ack_wait = Time.sec_to_nano(1.5)

        start_time = self.args.time
        if start_time:
            try:
                start_time = tz.to_str(start_time)
                self.logger.debug(f"Start time: {start_time}")
            except BadRequest as e:
                self.logger.error(str(e))
                return

        del_pol = ConsumerConfig.DeliverPolicy.BY_START_TIME if start_time else ConsumerConfig.DeliverPolicy.NEW

        try:
            if self.args.heartbeats:
                stream_tally += 1
                await self.inu.js.consumer.create(
                    Consumer(const.Streams.HEARTBEAT, ConsumerConfig(
                        filter_subject=const.Subjects.all(const.Subjects.HEARTBEAT),
                        deliver_policy=del_pol,
                        opt_start_time=start_time,
                        ack_wait=ack_wait,
                    )), self.on_hb,
                )

            if self.args.alerts:
                stream_tally += 1
                await self.inu.js.consumer.create(
                    Consumer(const.Streams.ALERTS, ConsumerConfig(
                        filter_subject=const.Subjects.all(const.Subjects.ALERT),
                        deliver_policy=del_pol,
                        opt_start_time=start_time,
                        ack_wait=ack_wait,
                    )), self.on_alert,
                )

            if self.args.settings:
                stream_tally += 1
                await self.inu.js.consumer.create(
                    Consumer(const.Streams.SETTINGS, ConsumerConfig(
                        filter_subject=const.Subjects.all(const.Subjects.SETTINGS),
                        deliver_policy=del_pol,
                        opt_start_time=start_time,
                        ack_wait=ack_wait,
                    )), self.on_settings,
                )

            if self.args.commands:
                stream_tally += 1
                await self.inu.js.consumer.create(
                    Consumer(const.Streams.COMMAND, ConsumerConfig(
                        filter_subject=const.Subjects.all(const.Subjects.COMMAND),
                        deliver_policy=del_pol,
                        opt_start_time=start_time,
                        ack_wait=ack_wait,
                    )), self.on_command,
                )

            if self.args.logs or stream_tally == 0:
                stream_tally += 1
                await self.inu.js.consumer.create(
                    Consumer(const.Streams.LOGS, ConsumerConfig(
                        filter_subject=const.Subjects.all(const.Subjects.LOG),
                        deliver_policy=del_pol,
                        opt_start_time=start_time,
                        ack_wait=ack_wait,
                    )), self.on_log,
                )

        except mn_error.NotFoundError:
            self.logger.error("Stream not found. Ensure NATS environment is bootstrapped.")
            return

        except ErrorResponseException as e:
            err = e.err_response
            self.logger.error(f"NATS: {err.code}-{err.err_code}: {err.description}")

        except Exception as e:
            self.logger.error(f"Subscribe error: {type(e).__name__}: {e}")
            return

        self.logger.info(f"Monitoring {stream_tally} stream{'s' if stream_tally > 0 else ''}..")

    async def handle_msg(self, msg: model.Message, msg_logic: callable):
        # Raw stream messages (eg last message) can't be ack'd
        if msg.can_ack():
            await self.inu.js.msg.ack(msg)

        if self.limit and self.logged >= self.limit:
            # overflow during shutdown
            return

        try:
            msg_logic()
        except error.Malformed as e:
            self.logger.error(f"Malformed message on subject '{msg.get_subject()}': {str(e)}")
        except Exception as e:
            self.logger.error(f"Error reading subject '{msg.get_subject()}': {str(e)}")

        if self.limit:
            self.logged += 1
            if self.logged >= self.limit:
                self.exit = True

    async def on_log(self, msg: model.Message):
        def handle_log():
            device_id = msg.get_subject()[len(const.Subjects.LOG) + 1:]
            log = Log(msg.get_payload())
            print(f"{msg.time} <{device_id}> [{log.level}] {log.message}")

        await self.handle_msg(msg, handle_log)

    async def on_alert(self, msg: model.Message):
        def handle_alert():
            device_id = msg.get_subject()[len(const.Subjects.ALERT) + 1:]
            alert = Alert(msg.get_payload())
            print(f"{msg.time} <{device_id}> [P{alert.priority}] {alert.message}")

        await self.handle_msg(msg, handle_alert)

    async def on_hb(self, msg: model.Message):
        def handle_hb():
            device_id = msg.get_subject()[len(const.Subjects.HEARTBEAT) + 1:]
            hb = Heartbeat(msg.get_payload())
            print(f"{msg.time} <{device_id}> BEAT ({hb.interval}s)")

        await self.handle_msg(msg, handle_hb)

    async def on_settings(self, msg: model.Message):
        def handle_hb():
            device_id = msg.get_subject()[len(const.Subjects.SETTINGS) + 1:]
            print(f"{msg.time} <{device_id}> NEW SETTINGS :: {msg.get_payload().decode()}")

        await self.handle_msg(msg, handle_hb)

    async def on_command(self, msg: model.Message):
        def handle_hb():
            device_id = msg.get_subject()[len(const.Subjects.COMMAND) + 1:]
            print(f"{msg.time} <{device_id}> CMD :: {msg.get_payload().decode()}")

        await self.handle_msg(msg, handle_hb)



from apa102 import Apa102 as LedStrip
from . import RoboticsDevice
from .control import Control
from .control.lights import Colour, Fx


class Apa102(RoboticsDevice):
    """
    APA102 LED strip.
    """
    CONFIG_ALIASES = ["apa102", "apa102c"]

    def __init__(self, num_leds: int, spi_index=1, segments=None, inu=None):
        """
        """
        super().__init__(inu=inu, log_path="inu.robotics.apa102")
        self.leds = LedStrip(num_leds, spi_index=spi_index)

        if segments is not None:
            for seg_id, (start, end) in segments.items():
                self.leds.create_segment(seg_id, start, end)

    async def execute(self, ctrl: Control, reverse: bool = False):
        await super().execute(ctrl)

        if isinstance(ctrl, Colour):
            # Fill the strip/segment fully & instantly with a single colour
            self.leds.fill(ctrl.colour, write=ctrl.execute)
        elif isinstance(ctrl, Fx):
            # Fx transitions -
            if ctrl.get_fx() == Fx.FX.FADE:
                # Full segment fade to colour
                self.leds.fade(ctrl.colour, ctrl.get_duration())
            elif ctrl.get_fx() == Fx.FX.SLIDE_L:
                # Slide "left"
                self.leds.slide(ctrl.colour, ctrl.get_duration(), direction=LedStrip.DIRECTION.LEFT)
            elif ctrl.get_fx() == Fx.FX.SLIDE_R:
                # Slide "right"
                self.leds.slide(ctrl.colour, ctrl.get_duration(), direction=LedStrip.DIRECTION.RIGHT)
            elif ctrl.get_fx() == Fx.FX.PULSE_L:
                # Pulse "left"
                self.leds.pulse(ctrl.colour, ctrl.get_duration(), direction=LedStrip.DIRECTION.LEFT)
            elif ctrl.get_fx() == Fx.FX.PULSE_R:
                # Pulse "right"
                self.leds.pulse(ctrl.colour, ctrl.get_duration(), direction=LedStrip.DIRECTION.RIGHT)

    def select_component(self, component_id):
        self.leds.select_segment(component_id)

    def __repr__(self):
        return f"APA102 strip"



import time

from machine import Pin


class SwitchMode:
    NO = "NO"
    NC = "NC"


class PullMode:
    PULL_DOWN = 1
    PULL_UP = 2


class Switch:
    def __init__(self, pin: int, pull: int = PullMode.PULL_DOWN, mode: str = SwitchMode.NO, min_active: int = 0,
                 on_change: callable = None):
        """
        If the `mode` is NC, then an active current on the pin will be considered "off".
        If `min_active` is set to non-zero, then the hardware must be in a new state for `min_active` ms before
        changing state.
        """
        self.pin = Pin(pin, Pin.IN, pull=pull)
        self.state = None
        self.on_change = on_change
        self.reversed = mode == SwitchMode.NC
        self.min_active = min_active

        # Time the hardware changed state, we'll time until `min_active` ms then trigger a change
        self.begin_state_change = None

        # Time the device has been in an active state
        self.active_time = None

    async def check_state(self, no_delay=False) -> bool:
        """
        Checks the state of the switch input pin. If the state has changed, `on_change(state)` will be called.

        Returns the switch state.
        """
        state = bool(self.pin.value())

        # Reverse the state if we're using a normally-closed circuit
        if self.reversed:
            state = not state

        if state != self.state:
            if not no_delay and self.min_active > 0:
                # We require a delay before changing state
                if self.begin_state_change is None:
                    self.begin_state_change = time.time()
                    return self.state
                else:
                    if time.time() - self.begin_state_change < (self.min_active / 1000):
                        return self.state

            # State has changed, start an active time and trigger callbacks
            self.state = state

            if state:
                self.active_time = time.time()
            else:
                self.active_time = None

            if self.on_change:
                await self.on_change(self.state)

        else:
            # State is the same, check if we should reset the state-change counter
            if self.begin_state_change is not None:
                # This will happen if the device physically changes state quickly in succession before the min active
                # time has expired (eg filter interference)
                self.begin_state_change = None

        return self.state

    def get_active_time(self):
        """
        Returns the time the device has been active for, in seconds.
        """
        if not self.active_time:
            return 0

        return time.time() - self.active_time



import asyncio
import json
import logging
import time

from micro_nats import io, model
from micro_nats.client import Client as NatsClient
from micro_nats.jetstream.client import Client as JsClient
from micro_nats.jetstream.error import ErrorResponseException
from micro_nats.jetstream.protocol import consumer
from micro_nats.util import Time
from micro_nats.util.asynchronous import TaskPool

from . import error, const, schema
from .schema import Heartbeat
from .schema.settings import Settings
from .schema.status import Status


class InuHandler:
    async def on_settings_updated(self):
        pass

    async def on_connect(self, server: model.ServerInfo):
        pass

    async def on_disconnect(self):
        pass


class Inu(io.IoHandler):
    def __init__(self, context: const.Context, handler: InuHandler | None = None):
        self.logger = logging.getLogger('inu')
        self.context = context
        self.settings = Settings()
        self.handler = handler
        self.pool = TaskPool()
        self.has_settings = False
        self.hb_task = None
        self.state = Status(enabled=False, locked=False, active=False, status="")

        # Update this with LAN address for heartbeats
        self.local_address = ""

        if isinstance(self.context.device_id, list):
            self.device_id = ".".join(self.context.device_id)
        else:
            self.device_id = self.context.device_id

        self.app_name = self.device_id.split(".")[0]

        if self.context.settings_class is not None and not issubclass(self.context.settings_class, Settings):
            raise error.BadRequest("Settings class is not a subclass of `Settings`")

        self.nats = NatsClient(model.ServerContext(self.context.nats_server), self)
        self.js = JsClient(self.nats)

    async def init(self) -> bool:
        """
        Brings an application online.

        Establishes a NATS connections, (if applicable) will acquire app settings, begins heartbeat, etc.

        Returns False if it could not get a NATS connection.
        """
        try:
            if not self.nats.is_connected():
                self.logger.debug("Init: NATS..")
                await self.nats.connect()
                await self._wait_for_nats_connect()

            return True

        except error.NoConnection:
            self.logger.error("No connection to NATS server, aborting")
            return False

        except Exception as e:
            self.logger.error(f"Init error: {type(e).__name__}: {e}")
            return False

    async def _wait_for_nats_connect(self):
        start_time = time.time()
        while not self.nats.is_connected():
            if time.time() - start_time > self.context.io_timeout:
                raise error.NoConnection("Timeout connecting to NATS server")

            await asyncio.sleep(0.01)

    async def init_settings(self):
        """
        Subscribes to the settings stream on the subject of this device id.

        `on_settings_updated()` will be called on the provided `NatsHandler` when new messages arrive.
        """
        assert self.context.settings_class is not None, "Settings class is not defined"

        try:
            self.logger.debug(f"Subscribing to settings against schema {self.context.settings_class.__name__}")
            await self.js.consumer.create(
                consumer.Consumer(
                    const.Streams.SETTINGS,
                    consumer_cfg=consumer.ConsumerConfig(
                        filter_subject=const.Subjects.fqs(const.Subjects.SETTINGS, self.device_id),
                        deliver_policy=consumer.ConsumerConfig.DeliverPolicy.LAST,
                        ack_wait=Time.sec_to_nano(5),
                    )
                ), push_callback=self._on_settings,
            )

        except ErrorResponseException as e:
            err = e.err_response
            self.logger.error(f"Error subscribing to device settings: {err.code}-{err.err_code}: {err.description}")

        except Exception as e:
            self.logger.error(f"Error subscribing to device settings: {type(e).__name__}: {e}")

    async def _on_settings(self, msg: model.Message):
        try:
            self.settings = self.context.settings_class(msg.get_payload())
            if self.has_settings:
                await self.log("Applied new settings")
            self.has_settings = True
            await self.js.msg.ack(msg)
        except Exception as e:
            await self.js.msg.nack(msg)
            await self.log(f"Error applying new settings: {type(e).__name__}: {e}")
            self.logger.error(f"Error updating device settings: {type(e).__name__}: {e}")

        if self.handler:
            self.pool.run(self.handler.on_settings_updated())

    async def init_heartbeat(self):
        """
        Creates a background worker that dispatches heartbeat messages at the configured interval.

        If the heartbeat interval changes (in the settings object), this worker will respond in kind.
        """
        if not self.settings.heartbeat_interval:
            self.settings.heartbeat_interval = 5

        self.logger.debug(f"Beginning heartbeat, interval {self.settings.heartbeat_interval}")
        self.hb_task = asyncio.create_task(self._run_heartbeat())

    async def _run_heartbeat(self):
        last_beat = 0

        try:
            while True:
                if not isinstance(self.settings, Settings) or not self.settings.heartbeat_interval:
                    interval = 5
                else:
                    interval = int(self.settings.heartbeat_interval)

                if time.time() - last_beat >= interval:
                    hb = Heartbeat(interval=interval, build=const.INU_BUILD, local_addr=self.local_address)

                    try:
                        self.logger.debug("beat")
                        await self.nats.publish(
                            const.Subjects.fqs(const.Subjects.HEARTBEAT, self.device_id), hb.marshal()
                        )
                    except Exception as e:
                        self.logger.error(f"Heartbeat error: {type(e).__name__}: {str(e)}")

                    last_beat = time.time()

                await asyncio.sleep(1)

        except asyncio.CancelledError:
            self.logger.debug(f"Heartbeat dying")

    async def _kill_heartbeat(self):
        if self.hb_task:
            self.hb_task.cancel()
            self.hb_task = None

    async def command(self, sub_cmd: str, data: dict = None):
        """
        Dispatch a command + sub-command message.
        """
        if data is None:
            data = {}

        try:
            await self.nats.publish(
                const.Subjects.fqs([const.Subjects.COMMAND, sub_cmd], self.device_id),
                json.dumps(data)
            )
        except Exception as e:
            self.logger.error(f"Command error: {type(e).__name__}: {str(e)}")

    async def status(self, active: bool = None, status: str = None, enabled: bool = None, locked: bool = None):
        """
        Update and publish the current device state.
        """
        if active is not None:
            self.state.active = active
        if status is not None:
            self.state.status = status
        if enabled is not None:
            self.state.enabled = enabled
        if locked is not None:
            self.state.locked = locked

        self.logger.debug(f"Device state: {self.state}")

        try:
            await self.nats.publish(
                const.Subjects.fqs(const.Subjects.STATUS, self.device_id),
                json.dumps(self.state.marshal())
            )
        except Exception as e:
            self.logger.error(f"Status error: {type(e).__name__}: {str(e)}")

    async def activate(self, status_msg: str = ""):
        """
        Sets state to active to True & sets the status string, then dispatches a `status` message.
        """
        await self.status(active=True, status=status_msg)

    async def deactivate(self, status_msg: str = ""):
        """
        Sets state active to False & clears the status string, then dispatches a `status` message.
        """
        await self.status(active=False, status=status_msg)

    async def log(self, message: str, level: str = const.LogLevel.INFO):
        """
        Dispatch a log message on the device's log subject.
        """
        await self.log_s(schema.Log(message=message, level=level))

    async def log_s(self, log: schema.Log):
        """
        Schema-form of `log()`
        """
        self.logger.info(f"LOG [{log.level.upper()}]: {log.message}")
        try:
            await self.nats.publish(const.Subjects.fqs(const.Subjects.LOG, self.device_id), log.marshal())
        except Exception as e:
            self.logger.error(f"Log error: {type(e).__name__}: {str(e)}")

    async def alert(self, message: str, priority: int = const.Priority.P3):
        """
        Create an alert, designed for human intervention.

        Typically, this will trigger an alarm - a pager for example - that will get immediate attention. Priority P1 &
        P2 are for urgent alerts (read: will wake user) and P3 & P4 for less urgent (read: will not wake).
        """
        await self.alert_s(schema.Alert(message=message, priority=priority))

    async def alert_s(self, alert: schema.Alert):
        """
        Schema-form of `alert()`.
        """
        self.logger.info(f"ALERT [P{alert.priority}]: {alert.message}")
        try:
            await self.nats.publish(const.Subjects.fqs(const.Subjects.ALERT, self.device_id), alert.marshal())
        except Exception as e:
            self.logger.error(f"Failed to publish alert: {type(e).__name__}: {str(e)}")

    async def on_connect(self, server: model.ServerInfo):
        if self.context.settings_class is not None:
            self.logger.debug("Init: settings..")
            await self.init_settings()

        if self.context.has_heartbeat:
            self.logger.debug("Init: heartbeat..")
            await self.init_heartbeat()

        self.logger.info(f"Connected to NATS server: {server}")
        self.pool.run(self.handler.on_connect(server))

    async def on_disconnect(self):
        await self.js.flush_inbox()
        await self._kill_heartbeat()
        self.has_settings = False
        self.pool.run(self.handler.on_disconnect())

    def get_central_id(self) -> str:
        return f"central.{self.device_id}"



INU_BUILD = 46


class LogLevel:
    DEBUG = "debug"
    INFO = "info"
    WARNING = "warning"
    ERROR = "error"
    FATAL = "fatal"


class Priority:
    P1 = 1
    P2 = 2
    P3 = 3
    P4 = 4
    P5 = 5

    HIGHEST = 1
    LOWEST = 5


class DeviceType:
    # Generic utility, such as a CLI, settings editor, bootstrap, etc
    UTILITY = "utility"

    # Monitor application, log viewer, etc
    MONITOR = "monitor"

    # On-off or timer switch - can be triggered by sensors, etc
    RELAY = "relay"

    # Physical switch(es)
    SWITCH = "switch"

    # Capacitive sensor
    CAPACITOR = "capacitor"

    # Motion, light, distance, etc sensors
    MOTION = "motion"
    RANGE = "range"
    LIGHT = "light"

    # Motorised actuator or physical pulley, winch, etc
    ROBOTICS = "robotics"


class Context:
    device_id: str | list[str] = None
    nats_server: str = None
    has_heartbeat: bool = False
    settings_class: type = None
    io_timeout: float = 3

    def __init__(self, device_id: str | list[str], **params):
        self.device_id = device_id

        for k, v in params.items():
            if k[0] == '_':
                continue
            setattr(self, k, v)


class Subjects:
    """
    Subject format should include the base subject + the device name:
        inu.log.some-sensor

    In addition, some subjects might have sub-subjects:
        inu.cmd.trigger.some-sensor

    Payloads are always JSON.
    """
    # Standard logging
    # '{'lvl': str, 'msg': str}
    # lvl: "debug", "info", "warning", "error", "fatal"
    LOG = 'log'

    # Alerts should raise human attention
    # {'priority': int8, 'msg': str}
    ALERT = 'alert'

    # Status updates, such as begin, end, etc
    # {'enabled': bool, 'active': bool, 'status': str}
    STATUS = 'status'

    # Base command, always requires a sub-command in the subject
    COMMAND = 'cmd'

    # Device has been activated, such as a button pressed or sensor trip
    # {'code': int}
    COMMAND_TRIGGER = 'trigger'

    # Jog a robotics actuator/servo - typically sent to the device's 'central' subject
    # {'device': str, 'distance': int, 'speed': int}
    COMMAND_JOG = 'jog'

    # Request a device perform an OTA update - must be sent to the device's 'central' subject
    # {'version': int}
    COMMAND_OTA = 'ota'

    # Request the device perform a soft or hard reboot
    # {}
    COMMAND_REBOOT = 'reboot'

    # Heartbeats let a controller know that you're still alive and detect devices going offline
    # Includes your build version and LAN IP address
    # {'uptime': int64, 'build': int, 'local_addr': str}
    HEARTBEAT = 'hb'

    # Settings define a device's configuration
    # {...}
    SETTINGS = 'settings'

    @staticmethod
    def all(subject: str | list[str], multi=True) -> str:
        """
        Get a wildcard subject.

        If `multi` is True, the wildcard will be >, otherwise it will use *.
        """
        wc = ">" if multi else "*"
        if isinstance(subject, list):
            subject = ".".join(subject)

        return ".".join([subject, wc])

    @staticmethod
    def fqs(subject: str | list[str], device: str | list[str]) -> str:
        """
        Get a fully-qualified subject string.
        """
        if isinstance(subject, list):
            subject = ".".join(subject)

        if isinstance(device, list):
            device = ".".join(device)

        return ".".join([subject, device])


class Streams:
    LOGS = 'logs'
    ALERTS = 'alerts'
    STATUS = 'status'
    COMMAND = 'commands'
    HEARTBEAT = 'heartbeats'
    SETTINGS = 'settings'


class TriggerCode:
    # [Robotics] Robotics interrupt - roll-back and restart
    INTERRUPT = 100
    # [Robotics] Request device recalibrate (must NOT be enabled)
    CALIBRATE = 101
    # Force-clear active state (repair lingering error without restarting)
    RESET_ACTIVE = 102
    # [Robotics] Similar to INTERRUPT, but will only reset the timer on an active WAIT command
    WAIT = 103
    # [Robotics] Will immediately drop out of an active WAIT command
    BREAK = 104

    ENABLE_TOGGLE = 110
    ENABLE_ON = 111
    ENABLE_OFF = 112

    LOCK_TOGGLE = 115
    LOCK_ON = 116
    LOCK_OFF = 117


class Strings:
    RANGE = "Range"
    SEQ = "Sequence"
    COOLDOWN = "Cooldown"



import re
import time
from datetime import datetime, date, timezone

from .error import BadRequest


def to_str(src: str) -> str:
    """
    Takes an input and converts it to a NATS-expected timestamp in the format "2023-09-13T10:17:59Z".

    Valid inputs relative to current time:
        4s
        3m4s
        20h3m4s
        82d20h3m4s

    Valid absolute inputs:
        2023-04-02
        2023-04-02T20:04
        2023-04-02T20:04:30
        2023-04-02T20:04:30.343

    Today-specific time:
        20:04
        20:04:30
        20:04:30.123

    Absolute formats MUST be in UTC. No time zone should be included.
    """
    src = src.replace(" ", "").upper()
    if src[-1:] == "Z":
        src = src[:-1]

    # Absolute dates
    if re.match("^[0-9]{4}-[0-9]{2}-[0-9]{2}$", src):
        return f"{src}T00:00:00Z"

    if re.match("^[0-9]{4}-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}$", src):
        return f"{src}:00Z"

    if re.match("^[0-9]{4}-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}:[0-9]{2}$", src):
        return f"{src}Z"

    if re.match("^[0-9]{4}-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}:[0-9]{2}\\.[0-9]{1,6}$", src):
        return f"{src}Z"

    # Time relative to today
    if re.match("^[0-9]{2}:[0-9]{2}$", src):
        return f"{date.today().strftime('%Y-%m-%d')}T{src}:00Z"

    if re.match("^[0-9]{2}:[0-9]{2}:[0-9]{2}$", src):
        return f"{date.today().strftime('%Y-%m-%d')}T{src}Z"

    if re.match("^[0-9]{2}:[0-9]{2}:[0-9]{2}\\.[0-9]{1,6}$", src):
        return f"{date.today().strftime('%Y-%m-%d')}T{src}Z"

    # Relative times
    cursor = time.time()
    history = []
    parts = re.findall(r"[0-9]+[A-Z]", src)

    recon = "".join(parts)
    if recon != src or len(parts) == 0:
        raise BadRequest(f"Invalid date format: {src}")

    for part in parts:
        mod = part[-1:]
        offset = int(part[:-1])

        if mod in history:
            raise BadRequest(f"Duplicate time modifier: {mod}")

        if not offset:
            raise BadRequest(f"Invalid time offset: {part[:-1]}")

        history.append(mod)

        if mod == "S":
            cursor = cursor - offset
        elif mod == "M":
            cursor = cursor - (offset * 60)
        elif mod == "H":
            cursor = cursor - (offset * 3600)
        elif mod == "D":
            cursor = cursor - (offset * 86400)

    if cursor == time.time():
        # not entirely sure how this can happen, but just in case -
        raise BadRequest(f"Invalid date format: {src}")

    return datetime.fromtimestamp(cursor, tz=timezone.utc).strftime("%Y-%m-%dT%H:%M:%SZ")



import time
from machine import I2C
from . import AmbientLightSensor


class VEML6030(AmbientLightSensor):
    # VEML6030 I2C address
    I2C_ADDR = 0x10

    # Configuration register
    ALS_CONF = 0x00  # Ambient Light Sensor (ALS) configuration register

    # lux/bit
    RESOLUTION = 0.0576

    class GAIN:
        GAIN_1 = 0x00  # Gain x1
        GAIN_2 = 0x01  # Gain x2
        GAIN_1_8 = 0x02  # Gain x1/8
        GAIN_1_4 = 0x03  # Gain x1/4

    def __init__(self, i2c_id=1, gain=GAIN.GAIN_1):
        self.i2c = I2C(i2c_id)
        self.config = bytearray(2)
        self.set_config(gain)  # Set ALS gain and integration time

    def read(self):
        # Read ALS data from 0x04 register
        data = self.i2c.readfrom_mem(self.I2C_ADDR, 0x04, 2)
        return ((data[0] << 8) | data[1]) * VEML6030.RESOLUTION

    def set_config(self, value):
        self.config[0] = value & 0xFF
        self.config[1] = (value >> 8) & 0xFF
        self.i2c.writeto_mem(self.I2C_ADDR, VEML6030.ALS_CONF, self.config)
        time.sleep(0.2)



from . import Control
from inu import error


class Select(Control):
    """
    Select the active device. Should precede execution controls like MV.
    """
    CONTROL_CODE = "SEL"
    ALIASES = ["SEL", "S", "SELECT"]

    def __init__(self, ctrl: str = None):
        super().__init__(ctrl)

        if self.code not in self.ALIASES or len(self.args) != 1:
            raise error.Malformed(f"Invalid {self.CONTROL_CODE} control: {ctrl}")

    def get_device(self) -> str:
        """
        Returns the selection subject ("XX" from "SEL XX:YY").
        """
        return self.args[0].split(":")[0]

    def get_component(self) -> str:
        """
        Returns the selection component ("TT" from "SEL XX:YY").
        """
        return self.args[0].split(":")[1] if ":" in self.args[0] else None

    def allow_interrupt(self) -> bool:
        """
        Select Controls have no concept of interrupts but should not block the interrupt chain.
        """
        return True

    def __repr__(self):
        return f"SEL {self.get_device()}"


class Wait(Control):
    """
    Delay by a given time in milliseconds.

    The time can also be specified in seconds (s), minutes (m) or hours (h) by using the respective suffix.
    """
    CONTROL_CODE = "WAIT"
    ALIASES = ["W", "WAIT"]

    def __init__(self, ctrl: str = None):
        super().__init__(ctrl)

        if self.code not in self.ALIASES or len(self.args) != 1:
            raise error.Malformed(f"Invalid {self.CONTROL_CODE} control: {ctrl}")

        self.time = 0
        spec = self.args[0].upper().strip()

        if spec.endswith("S"):
            self.time = int(spec[:-1]) * 1000
        elif spec.endswith("M"):
            self.time = int(spec[:-1]) * 60 * 1000
        elif spec.endswith("H"):
            self.time = int(spec[:-1]) * 60 * 60 * 1000
        else:
            self.time = int(spec)

    def get_time(self) -> int:
        """
        Wait time as an integer in milliseconds.
        """
        return self.time

    def __repr__(self):
        return f"WAIT {self.get_time()}"


class Trigger(Control):
    """
    Send a Trigger code.
    """
    CONTROL_CODE = "TRG"
    ALIASES = ["TRG", "TRIGGER"]

    def __init__(self, ctrl: str = None):
        super().__init__(ctrl)

        if self.code not in self.ALIASES or len(self.args) != 1:
            raise error.Malformed(f"Invalid {self.CONTROL_CODE} control: {ctrl}")

    def get_code(self) -> int:
        """
        Trigger code as an integer.
        """
        return int(self.args[0])

    def __repr__(self):
        return f"TRG {self.get_code()}"



import time

from inu import const
from inu.app import InuApp
from inu.hardware.switch import Switch, SwitchMode


class SwitchManager(InuApp):
    def __init__(self, settings_class: type):
        super().__init__(settings_class)
        self.switches = []
        self.switches_active = 0
        self.activate_on_switch = True
        self.fallback_refire_delay = None

    async def switch_init(self):
        def device_cfg(path, keys, default=None):
            for key in keys:
                if key not in path:
                    return default
                path = path[key]
            return path

        self.fallback_refire_delay = self.get_config(["switch", "refire_delay"], None)

        index = 0
        devices = self.get_config(["switch", "devices"], [])
        for device in devices:
            mode = device_cfg(device, ["mode"], SwitchMode.NO)
            pin = device_cfg(device, ["pin"], None)
            name = device_cfg(device, ["name"], f"s{index}")
            code = device_cfg(device, ["code"], None)
            if not pin:
                self.logger.error("No pin assigned for switch")
                continue

            self.switches.append((Switch(pin=pin, mode=mode), name, code))
            index += 1

    async def switch_tick(self):
        active = 0

        if hasattr(self.inu.settings, "refire_delay"):
            refire_delay = self.inu.settings.refire_delay
        else:
            refire_delay = self.fallback_refire_delay

        for i, (sw, name, code) in enumerate(self.switches):
            # Check each switch for state change
            last_state = sw.state

            # Force all devices to be considered "off" if we've disabled the device
            if self.inu.state.enabled:
                if self.inu.state.locked:
                    # Locked - don't change state (except for disabling)
                    new_state = last_state
                else:
                    new_state = await sw.check_state()
            else:
                new_state = False
                sw.state = False

            if new_state:
                active += 1

                if refire_delay and (sw.get_active_time() >= (refire_delay / 1000)):
                    # Send a re-fire trigger
                    await self.fire_trigger(name, self.get_code_for_switch(i, code))
                    sw.active_time = time.time()

            if last_state != new_state:
                # State changed
                self.logger.info(f"Switch '{name}': {last_state} -> {new_state}")

                if new_state:
                    # If we're moving into an active state, gather the right code and fire -
                    await self.fire_trigger(name, self.get_code_for_switch(i, code))

        if active != self.switches_active:
            # Update the device status if the number of active switches changed
            self.switches_active = active

            if self.activate_on_switch:
                if active > 0:
                    await self.inu.activate(f"Active: {active}")
                else:
                    await self.inu.deactivate()

    def get_code_for_switch(self, index: int, code=None):
        if index > 5:
            # There are only 6 override codes
            code = code or -1
        else:
            if hasattr(self.inu.settings, f"sw_{index}"):
                code = int(getattr(self.inu.settings, f"sw_{index}"))
            else:
                code = code or -1

        # If override code is -1, use the default code (trigger_code)
        if code == -1:
            if hasattr(self.inu.settings, "trigger_code"):
                code = int(self.inu.settings.trigger_code)
            else:
                code = 0

        return code

    async def fire_trigger(self, name, code):
        self.logger.info(f"Switch '{name}' firing; code {code}")
        await self.inu.command(const.Subjects.COMMAND_TRIGGER, {
            'code': code,
        })



import argparse
import binascii
import glob
import json
import logging
import os
import re
import shutil
import struct
import tarfile
import tempfile
from os.path import dirname as up

import aiofiles
import aiohttp
from gcloud import storage
from mpremote import commands as mpremote
from mpremote.main import State

from inu import const
from inu.util import Utility


class MpArgs:
    verbose = False
    recursive = False

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)


class File:
    def __init__(self, abs_fn: str, short_fn: str, is_dir: bool):
        self.abs_fn = abs_fn
        self.short_fn = short_fn
        self.is_dir = is_dir


class Build(Utility):
    # Github repo details for the MicroPython library
    LIB_REPO = "micropython/micropython-lib"
    LIB_TAG = "v1.21.0"

    # Github archive structure
    GIT_PATH = "https://github.com/{lib}/archive/refs/tags/{tag}{ext}"

    # Over-the-air updates
    OTA_BUCKET = "inu-ota"

    def __init__(self, args: argparse.Namespace):
        super().__init__(args)
        self.logger = logging.getLogger('inu.util.build')
        self.root_dir = up(up(up(up(os.path.realpath(__file__)))))
        self.local_dir = None
        self.ota_packet = None
        self.state = None

        # Set to True if you want to include config in the OTA update
        self.force_include_cfg_ota = False

    async def run(self):
        """
        Run the device builder.
        """
        device_id = self.args.device_id[0].split(".")

        if not os.path.exists(f"{self.root_dir}/apps/{device_id[0]}"):
            self.logger.error(f"App files for '{device_id[0]}' do not exist")
            exit(1)

        if self.force_include_cfg_ota or not self.args.ota:
            cfg = self.create_config_file(device_id)
            print_cfg = re.sub(r'"password": "(.*)"', '"password": "********"', cfg)
            self.logger.info(f"Generated configuration:\n{print_cfg}")
        else:
            cfg = None
            self.logger.info(f"Generating OTA package for {device_id[0]} v{const.INU_BUILD}")

        port = self.args.port
        self.state = State()
        await self.get_libs(self.LIB_TAG)

        if self.args.local or self.args.ota:
            if self.args.settings:
                self.logger.error("You cannot use --settings with --ota or --local")
                return

            self.local_dir = self.root_dir + f"/build/apps/{'.'.join(device_id)}/"
            if os.path.exists(self.local_dir):
                shutil.rmtree(self.local_dir)
            os.makedirs(self.local_dir, exist_ok=True)

        if self.args.ota:
            self.ota_packet = struct.pack("<I", const.INU_BUILD)

        self.logger.info(f"Building {device_id}..")

        if self.local_dir:
            self.logger.info(f"Building to local directory: {self.local_dir}")
        else:
            self.logger.info(f"Connecting to device on {port}..")
            mpremote.do_connect(self.state, MpArgs(device=[port]))

        self.logger.info("Sending content..")

        # Core app
        self.send_files(f"apps/{device_id[0]}")

        if not self.args.ota:
            # Don't include the settings file in OTA updates
            with tempfile.NamedTemporaryFile('w') as fp:
                fp.write(cfg)
                fp.seek(0)
                self.send_file(fp.name, "settings.json")

        if self.args.settings:
            self.logger.info("New settings applied")
            return

        self.send_files("src/inu", "inu")
        self.send_files("src/micro_nats", "micro_nats")
        self.send_files("src/wifi", "wifi")
        self.send_files("src/apa102", "apa102")
        self.send_files("src/util", "util")

        # MicroPython libs
        if not self.args.ota:
            self.mkdir("lib")

        mc = self.get_lib_root(short_form=True)
        self.send_file(f"{mc}/python-stdlib/base64/base64.py", "lib/base64.py")
        self.send_file(f"{mc}/python-stdlib/datetime/datetime.py", "lib/datetime.py")
        self.send_file(f"{mc}/python-stdlib/logging/logging.py", "lib/logging.py")

        if not self.local_dir:
            self.logger.debug("Disconnecting..")
            mpremote.do_disconnect(self.state, MpArgs())

        if self.ota_packet:
            self.logger.info("Generating package checksum..")
            checksum = "%08X" % (binascii.crc32(self.ota_packet) & 0xFFFFFFFF)

            self.logger.info("Transmitting OTA package..")
            with open(self.local_dir + f"build-{const.INU_BUILD}.ota", "wb") as fp:
                fp.write(self.ota_packet)

            self.send_ota_package(device_id[0], checksum)

        self.logger.info("Done")

    def send_ota_package(self, app_code: str, checksum: str):
        client = storage.Client.from_service_account_json(f"{self.root_dir}/config/ota-service-account.json")
        bucket = client.get_bucket(self.OTA_BUCKET)

        ota_object = bucket.blob(f"{app_code}/build-{const.INU_BUILD}.ota")
        ota_object.cache_control = None
        ota_object.upload_from_filename(self.local_dir + f"build-{const.INU_BUILD}.ota")

        checksum_object = bucket.blob(f"{app_code}/build-{const.INU_BUILD}.checksum")
        checksum_object.cache_control = None
        checksum_object.upload_from_string(checksum)

        version_object = bucket.blob(f"{app_code}/version")
        version_object.cache_control = None
        version_object.upload_from_string(str(const.INU_BUILD).encode())

    def create_config_file(self, device_id) -> str:
        """
        Merges all base & device config files into one and returns them as a string (containing valid JSON).
        """
        dvc_id_str = '.'.join(device_id)
        files = [
            f"{self.root_dir}/config/base.json",
            f"{self.root_dir}/config/{device_id[0]}.json",
        ]

        fn = device_id[0]
        for part in device_id[1:]:
            fn += f".{part}"
            files.append(f"{self.root_dir}/config/{fn}.json")

        def merge(source, destination):
            for key, value in source.items():
                if isinstance(value, dict):
                    # get node or create one
                    node = destination.setdefault(key, {})
                    merge(value, node)
                else:
                    destination[key] = value

            return destination

        out = {'device_id': dvc_id_str, 'hostname': '_'.join(device_id)}
        for file in files:
            if os.path.exists(file):
                self.logger.info(f"Loading device info from {file}")
                try:
                    with open(file, 'r') as fp:
                        merge(json.load(fp), out)
                except Exception as e:
                    self.logger.error(f"ERROR loading config: {type(e).__name__}: {e}")

        return json.dumps(out, indent=4)

    def has_lib_files(self) -> bool:
        """
        Checks if we have a copy of the MicroPython libraries downloaded.
        """
        return os.path.exists(self.get_lib_root())

    def get_lib_root(self, short_form: bool = False) -> str:
        """
        Returns the path to where the MicroPython libraries should be.
        """
        prefix = "build/" if short_form else f"{self.root_dir}/build/"
        return f"{prefix}{self.LIB_REPO.split('/')[1]}-{self.LIB_TAG[1:]}"

    def get_files(self, path) -> list[File]:
        """
        Get a list of files in a given local directory.
        """
        files = []
        base_len = len(self.root_dir + f'/{path}/')
        listing = glob.glob(self.root_dir + f'/{path}/**/*', recursive=True)
        for line in listing:
            if "__pycache__" in line[base_len:]:
                continue
            files.append(File(abs_fn=line, short_fn=line[base_len:], is_dir=os.path.isdir(line)))

        return files

    def mkdir(self, path: str) -> bool:
        """
        Create a directory on a device connected via serial.
        """
        self.logger.debug(f"mkdir {path}")

        if self.local_dir:
            os.makedirs(self.local_dir + path, exist_ok=True)
            return True

        try:
            self.state.ensure_raw_repl()
            self.state.did_action()
            self.state.transport.fs_mkdir(path)
            return True
        except mpremote.TransportError as e:
            if "[Errno 17] EEXIST" in str(e):
                return False
            else:
                raise e

    def mkdir_path(self, path: str):
        """
        Create a directory structure on target filesystem.
        """
        base_paths = path.split("/")
        base = ""
        for p in base_paths:
            if base:
                base += f"/{p}"
            else:
                base = p

            self.mkdir(f"{base}")

    def send_files(self, src: str, dest: str = None):
        """
        Send files in a local directory to the target directory.

        Will automatically create required directory structure.
        """
        if dest is None:
            dest = ""
        else:
            if self.ota_packet is None:
                self.mkdir_path(dest)
            dest += "/"

        files = self.get_files(src)
        for file in files:
            tgt = f"{dest}{file.short_fn}"
            if file.is_dir:
                if self.ota_packet is None:
                    self.mkdir(tgt)
            else:
                self.logger.debug(f"cp {tgt}")

                if self.ota_packet:
                    self.add_ota_file(file.abs_fn, tgt)
                elif self.local_dir:
                    shutil.copy(file.abs_fn, self.local_dir + tgt)
                else:
                    mpremote.do_filesystem(self.state, MpArgs(command=["cp"], path=[file.abs_fn, f":{tgt}"]))

    def send_file(self, src: str, dest: str):
        """
        Send a single file to a remote device connected via serial.

        `dest` should be a file, and the base directory must already exist.
        """
        if src[0] != "/":
            src = f"{self.root_dir}/{src}"

        self.logger.debug(f"cp {dest}")
        if self.ota_packet:
            self.add_ota_file(src, dest)
        elif self.local_dir:
            shutil.copy(src, self.local_dir + dest)
        else:
            mpremote.do_filesystem(self.state, MpArgs(command=["cp"], path=[src, f":{dest}"]))

    def add_ota_file(self, src: str, dest: str):
        with open(src, "rb") as fp:
            content = fp.read()
            self.ota_packet += struct.pack("<H", len(dest))
            self.ota_packet += dest.encode()
            self.ota_packet += struct.pack("<I", len(content))
            self.ota_packet += content

    async def get_libs(self, tag: str, ext: str = ".tar.gz"):
        """
        Grabs a given release archive from Github and extracts it to the build directory. Does nothing if required
        files already exist.
        """
        url = self.GIT_PATH.format(lib=self.LIB_REPO, tag=tag, ext=ext)
        local_fn = f"{self.LIB_REPO.replace('/', '_')}-{tag}{ext}"
        local_path = f"{self.root_dir}/build/{local_fn}"

        if not os.path.exists(local_path):
            # Download release archive from Github
            self.logger.info(f"Downloading library archive {tag}..")
            async with aiohttp.ClientSession() as session:
                async with session.get(url) as response:
                    self.logger.debug(f"GET {url}: {response.status}")

                    resp = await response.read()

            # Write to local filesystem
            self.logger.debug(f"Writing to {local_path}..")
            async with aiofiles.open(local_path, "wb") as out:
                await out.write(resp)
                await out.flush()

        # Extract tarball
        if not self.has_lib_files():
            self.logger.debug(f"Extracting {local_fn}..")
            with tarfile.open(local_path, 'r') as tar:
                for item in tar:
                    tar.extract(item, f"{self.root_dir}/build/")



from . import Schema


class Command(Schema):
    pass


class Trigger(Command):
    code: int = None


class Jog(Command):
    device_id: str = None
    distance: int = 0
    speed: int = 0


class Ota(Command):
    version: int = None


class Reboot(Command):
    pass



from . import Settings, CooldownDevice, ActionDevice


class Robotics(Settings, ActionDevice, CooldownDevice):
    """
    ## A collection of 1 or more actuators, steppers or servos that control a robotic action.

    **Control Codes**
        ```
        SEL <DEVICE ID>               Select a device
        MV <DISTANCE> <SPEED> [INT]   Move DISTANCE mm, at speed of SPEED mm/s
        W <TIME> [INT]                Pause for TIME milliseconds
        ```

    If "INT" is appended to the command, it will allow an interrupt trigger (code 100) to reset the operation. INT
    operations may be chained, the full chain will be reversed and eventually replayed.

    Example:
        SEL A0; MV 800 300; W 2000 INT; MV -800 150 INT

        Selects device "A0"
        Move actuator 800 mm at 300 mm/s
        Wait 2 seconds
        Move in reverse direction 800 mm at 150 mm/s
    """
    cal_seq: str = ""
    cal_seq_hint: str = "Calibration sequence when first initialising"

    trigger_wait: bool = False
    trigger_wait_hint: str = "Triggers during a sequence will count as 'wait' trigger (code 103: reset WAIT timer)"

    idle_power: bool = True
    idle_power_hint: str = "Keep devices powered when robotics is idle"

    idle_period: int = 900
    idle_period_hint: str = "Inactivity period (s) before a device is considered idle"

    warmup_delay: int = 2500
    warmup_delay_hint: str = "Time to wait (ms) after powering up devices before starting a sequence"

    seq_0: str = ""
    seq_0_hint: str = "Sequence 0 control codes"

    seq_1: str = ""
    seq_1_hint: str = "Sequence 1 control codes"

    seq_2: str = ""
    seq_2_hint: str = "Sequence 2 control codes"

    seq_3: str = ""
    seq_3_hint: str = "Sequence 3 control codes"

    seq_4: str = ""
    seq_4_hint: str = "Sequence 4 control codes"

    seq_5: str = ""
    seq_5_hint: str = "Sequence 5 control codes"



import argparse
import asyncio
import logging
import random
import sys
import time

from inu import Inu, InuHandler, const, Status
from inu import error
from inu.const import Context
from inu.schema.command import Trigger
from inu.schema.settings import Settings
from inu.util import Utility
from micro_nats import model
from micro_nats.jetstream.protocol.consumer import Consumer, ConsumerConfig
from micro_nats.util import Time


class ModeToggle(Utility, InuHandler):
    class Mode:
        LOCK = 1
        UNLOCK = 2
        ENABLE = 3
        DISABLE = 4

    ICO_ENABLED = "\U0001F7E2 Enabled"
    ICO_DISABLED = "\U0001F534 DISABLED"
    ICO_LOCKED = "\U0001F534 LOCKED"
    ICO_UNLOCKED = "\U0001F7E2 Unlocked"

    def __init__(self, args: argparse.Namespace, mode):
        super().__init__(args)
        self.logger = logging.getLogger('inu.util.toggle')
        self.inu = Inu(Context(
            device_id=["toggle", f"i{random.randint(1000, 9999)}"],
            nats_server=args.nats,
            has_heartbeat=False,
            settings_class=Settings,
        ), self)
        self.mode = mode
        self.device_id = args.device_id[0]

    async def run(self):
        """
        """
        if not await self.inu.init():
            return

        status = None

        async def on_status(msg: model.Message):
            nonlocal status
            if status is None:
                status = Status(msg.get_payload())
            else:
                status = Status(msg.get_payload())
                print(f"Enabled:  {self.ICO_ENABLED if status.enabled else self.ICO_DISABLED}")
                print(f"Lock:     {self.ICO_LOCKED if status.locked else self.ICO_UNLOCKED}")
                print(f"Status:   {status.status}")

        await self.inu.js.consumer.create(
            Consumer(const.Streams.STATUS, ConsumerConfig(
                filter_subject=const.Subjects.fqs(const.Subjects.STATUS, self.device_id),
                deliver_policy=ConsumerConfig.DeliverPolicy.LAST,
                ack_wait=Time.sec_to_nano(1.5),
            )), on_status,
        )

        # Wait for last known state
        start = time.monotonic()
        while status is None:
            if time.monotonic() - start > 1:
                self.logger.warning("No last known state")
                break

            await asyncio.sleep(0.01)

        payload = Trigger()
        if self.mode == self.Mode.LOCK:
            if isinstance(status, Status) and status.locked:
                print("Device already locked")
                sys.exit(2)
            payload.code = const.TriggerCode.LOCK_ON
        elif self.mode == self.Mode.UNLOCK:
            if isinstance(status, Status) and not status.locked:
                print("Device already unlocked")
                sys.exit(2)
            payload.code = const.TriggerCode.LOCK_OFF
        elif self.mode == self.Mode.ENABLE:
            if isinstance(status, Status) and status.enabled:
                print("Device already enabled")
                sys.exit(2)
            payload.code = const.TriggerCode.ENABLE_ON
        elif self.mode == self.Mode.DISABLE:
            if isinstance(status, Status) and not status.enabled:
                print("Device already disabled")
                sys.exit(2)
            payload.code = const.TriggerCode.ENABLE_OFF
        else:
            raise error.BadRequest(f"Bad toggle mode: {self.mode}")

        try:
            status = False
            self.logger.info(f"Sending status command to {self.device_id}..")
            await self.inu.nats.publish(
                const.Subjects.fqs(
                    [const.Subjects.COMMAND, const.Subjects.COMMAND_TRIGGER],
                    f"central.{self.device_id}"
                ),
                payload.marshal()
            )
        except Exception as e:
            self.logger.error(f"Publish error: {type(e).__name__}: {str(e)}")

        # Wait for status reply
        start = time.monotonic()
        while status is False:
            if time.monotonic() - start > 3:
                self.logger.error("Timeout waiting for device update")
                break

            await asyncio.sleep(0.01)



class RangingSensor:
    async def read_loop(self):
        """
        Starts the sensor reading loop.
        """
        pass

    def get_distance(self) -> int | None:
        """
        Returns the distance in mm.

        Returns None if no measurement has been made.
        """
        pass

    def get_age(self) -> float | None:
        """
        Returns the number of seconds since the last valid measurement.

        Returns None if no measurement has been made.
        """
        pass


class MotionSensor:
    async def read_loop(self):
        """
        Starts the sensor reading loop.
        """
        pass

    def is_motion(self) -> bool:
        """
        Returns the distance in mm.

        Returns None if no measurement has been made.
        """
        pass



from .. import Schema
from ... import error
from ...const import DeviceType, Priority


class Settings(Schema):
    """
    Base class for all settings.
    """
    heartbeat_interval: int = 5
    heartbeat_interval_hint: str = "Time in seconds to broadcast a device heartbeat (1 to 60)"
    heartbeat_interval_min: int = 1
    heartbeat_interval_max: int = 60

    device_priority: int = 3
    device_priority_hint: str = "Modifies the priority of alerts; 1: highest, 4: low, 5: does not page"
    device_priority_min: int = Priority.HIGHEST
    device_priority_max: int = Priority.LOWEST


class CooldownDevice:
    """
    Device has a cool-down period after execution.
    """
    cooldown_time: int = 5000
    cooldown_time_hint: str = "Time in ms before the device will reactivate"
    cooldown_time_min: int = 0


class ActionDevice:
    """
    Device listens to commands from other devices.
    """
    listen_subjects: str = ""
    listen_subjects_hint: str = "Will respond to messages (such as triggers) from subjects matching these strings; " \
                                "space delimited"


class Relay(Settings, ActionDevice):
    """
    ## An on-off or timer based power relay.

    Will either toggle, time-delay activate or power on/off a relay depending on the trigger code sent:
     0: Will toggle if `time_delay` is zero, or time-delay activate if non-zero
     1: Activate the relay
     2: Deactivate the relay
    """
    time_delay: int = 30
    time_delay_hint: str = "Auto-off time in seconds; if 0, device will be in pure toggle mode"
    time_delay_min: int = 0


def get_device_type_from_id(device_id: str) -> str:
    """
    Returns the device type from a device ID.

    "motion.foo" -> "motion"
    """
    parts = device_id.split(".")
    if len(parts) < 2:
        raise error.InvalidDeviceId()

    device_type = parts[0].strip().lower()
    # validate device type -
    get_device_settings_class(device_type)

    return device_type


def get_device_settings_class(dvc: str) -> type:
    """
    Find the correct settings class for the given device type.
    """
    from . import sensors, robotics

    if dvc == DeviceType.MOTION:
        return sensors.MotionSensor
    elif dvc == DeviceType.RANGE:
        return sensors.RangeTrigger
    elif dvc == DeviceType.ROBOTICS:
        return robotics.Robotics
    elif dvc == DeviceType.RELAY:
        return Relay
    elif dvc == DeviceType.SWITCH:
        return sensors.MultiSwitch
    elif dvc == DeviceType.CAPACITOR:
        return sensors.Capacitor
    elif dvc == DeviceType.LIGHT:
        return sensors.LightSensor
    else:
        raise error.UnsupportedDeviceType(f"Unsupported device type: {dvc}")


def get_hint_for_settings_class(settings_class) -> str:
    hint = ""
    doc = settings_class.__doc__ or ""
    for line in doc.strip().split("\n"):
        line = line.strip()

        if len(hint) > 0:
            hint += "\n" + line.strip()
        else:
            hint = line.strip()

    return hint


def get_config_for_settings_class(settings_class) -> dict[str, (str, str, int | None, int | None)]:
    config = {}

    def add_class(cls: type):
        if '__annotations__' not in cls.__dict__:
            return

        for base in cls.__bases__:
            add_class(base)

        for k, v in cls.__dict__['__annotations__'].items():
            if k[-5:] == "_hint" or k[-4:] == "_min" or k[-4:] == "_max":
                continue

            if f"{k}_hint" in cls.__dict__:
                hint = cls.__dict__[f"{k}_hint"]
            else:
                hint = ""

            if f"{k}_min" in cls.__dict__:
                conf_min = cls.__dict__[f"{k}_min"]
            else:
                conf_min = None

            if f"{k}_max" in cls.__dict__:
                conf_max = cls.__dict__[f"{k}_max"]
            else:
                conf_max = None

            config[k] = (v.__name__, hint, conf_min, conf_max)

    add_class(settings_class)

    return config


def get_config_for_device(device_type: str) -> (str, dict[str, (str, str, int | None, int | None)]):
    """
    Returns both the device type hint and a dictionary of settings.
    """
    cls = get_device_settings_class(device_type)
    return get_hint_for_settings_class(cls), get_config_for_settings_class(cls)



import asyncio
import json
import logging

import machine
from inu import InuHandler, Inu, const, Status
from inu.const import LogLevel
from inu.schema.command import Ota, Trigger, Reboot
from inu.updater import OtaUpdater
from micro_nats import model
from micro_nats.error import NotFoundError
from micro_nats.jetstream.error import ErrorResponseException
from micro_nats.jetstream.protocol import consumer
from micro_nats.util import Time
from wifi import Wifi, error as wifi_err


class InuApp(InuHandler):
    def __init__(self, settings_class: type):
        self.config = {}
        self.wifi = Wifi()
        self.load_config()
        self.listen_device_consumers = []

        # Set to false during device maintenance (such as an OTA update)
        self.allow_app_tick = True

        log_level = self.get_config('log_level', 'INFO')
        if not hasattr(logging, log_level):
            print(f"Invalid log level: {log_level}")
            exit(1)

        logging.basicConfig(level=getattr(logging, log_level))
        self.logger = logging.getLogger('app')

        self.inu = Inu(const.Context(
            device_id=self.get_config('device_id').split('.'),
            nats_server=self.get_config(['nats', 'server']),
            has_heartbeat=self.get_config('heartbeat', True),
            settings_class=settings_class,
        ), self)

    def load_config(self):
        """
        Load the settings.json file. Should only ever be called once.
        """
        with open("settings.json") as fp:
            self.config = json.load(fp)

        if 'wifi' in self.config:
            wifi = self.config['wifi']
            if 'ssid' in wifi and 'password' in wifi:
                self.wifi.set_network(wifi['ssid'], wifi['password'])
            else:
                print("Keys 'wifi' and/or 'password' missing from wifi configuration")
        else:
            print("Wifi configuration missing from settings")

    def get_config(self, key: str | list, default=None):
        """
        Get a setting from the local config, or return the default.
        """
        if isinstance(key, list):
            path = self.config
            for subkey in key:
                if subkey not in path:
                    return default
                path = path[subkey]
            return path

        if key in self.config:
            return self.config[key]
        else:
            return default

    async def connect_wifi(self) -> bool:
        try:
            self.wifi.connect()
        except OSError as e:
            self.logger.info(f"Wifi error: {e} -- delay retry")
            await asyncio.sleep(3)
            return False

        try:
            await self.wifi.wait_for_connect()
            self.logger.info("Wifi connected")
            return True
        except wifi_err.BadPassword:
            self.logger.error("Wifi password incorrect")
        except wifi_err.NoAccessPoint:
            self.logger.error(f"No access point responding to {self.wifi.ssid}")
        except wifi_err.Timeout:
            self.logger.error("Timeout connecting wifi")
        except wifi_err.WifiError as e:
            self.logger.error(f"Wifi connection error: {e}")

        return False

    async def init(self) -> bool:
        """
        Init the application. Executed by `run()`.

        Returns false if a key init item fails.
        """
        print("-- INU DEVICE STARTING --")
        print(f"Device ID: {self.inu.device_id}\n")

        print("Starting wifi..")
        if not await self.connect_wifi():
            print("Wifi connection failed, exiting")
            return False

        ifcfg = self.wifi.ifconfig()
        self.inu.local_address = ifcfg.ip
        print(f"  IP:      {ifcfg.ip}")
        print(f"  Subnet:  {ifcfg.subnet}")
        print(f"  Gateway: {ifcfg.gateway}")
        print(f"  DNS:     {ifcfg.dns}")

        print(f"\nBringing up NATS on {self.inu.context.nats_server}..")
        if not await self.inu.init():
            print("NATS connection failure, exiting")
            return False

        print("Waiting for settings..")
        while not self.inu.has_settings:
            await asyncio.sleep(0.1)

        print("\nBootstrap complete\n")
        await self.inu.log(f"ONLINE // Inu build {const.INU_BUILD} at {ifcfg.ip}")

        return True

    async def app_init(self):
        """
        Called once when the device boots. Override.
        """
        pass

    async def run(self):
        """
        Indefinite app loop.

        Checkup on wifi, NATS connection, etc. Will call `app_tick()` inside the main loop.
        """
        try:
            if not await self.init():
                self.logger.error("Init failed. Rebooting.")
                await asyncio.sleep(1)
                machine.reset()

            await self.app_init()

            while True:
                if not self.wifi.is_connected():
                    self.wifi.connect()
                    await self.wifi.wait_for_connect()

                    if not self.wifi.is_connected():
                        # If we have persistent wifi issues, do a full reboot
                        machine.reset()

                if self.allow_app_tick:
                    try:
                        await self.app_tick()
                    except Exception as e:
                        await self.inu.log(f"Application error - {type(e).__name__}: {e}", LogLevel.ERROR)
                        await asyncio.sleep(1)

                await asyncio.sleep(0.01)

        finally:
            # Reset on uncaught exception
            machine.reset()

    async def app_tick(self):
        """
        Override this with your application-specific logic. Called inside main_loop().
        """
        pass

    async def parse_trigger_code(self, code: int, msg):
        """
        Appropriately process a trigger message for the given code.

        This will execute special logic for the special codes (const.TriggerCode codes). Will pass on to `on_trigger()`
        if the code is non-special
        """
        if code == const.TriggerCode.INTERRUPT:
            await self.on_interrupt()
        elif code == const.TriggerCode.WAIT:
            await self.on_wait()
        elif code == const.TriggerCode.BREAK:
            await self.on_break()
        elif code == const.TriggerCode.RESET_ACTIVE:
            await self.inu.log(f"Indiscriminately resetting active state by user request", LogLevel.WARNING)
            await self.inu.status(active=False, status="")
        elif code == const.TriggerCode.ENABLE_TOGGLE:
            await self.inu.status(enabled=not self.inu.state.enabled, status="")
            await self.on_enabled_changed(self.inu.state.enabled)
        elif code == const.TriggerCode.ENABLE_ON:
            if not self.inu.state.enabled:
                await self.inu.status(enabled=True, status="")
                await self.on_enabled_changed(True)
        elif code == const.TriggerCode.ENABLE_OFF:
            if self.inu.state.enabled:
                await self.inu.status(enabled=False, status="")
                await self.on_enabled_changed(False)
        elif code == const.TriggerCode.LOCK_TOGGLE:
            await self.inu.status(locked=not self.inu.state.locked)
            await self.on_lock_changed(self.inu.state.locked)
        elif code == const.TriggerCode.LOCK_ON:
            if not self.inu.state.locked:
                await self.inu.status(locked=True)
                await self.on_lock_changed(True)
        elif code == const.TriggerCode.LOCK_OFF:
            if self.inu.state.locked:
                await self.inu.status(locked=False)
                await self.on_lock_changed(False)
        else:
            # Normal triggers can only execute when enabled
            await self.on_trigger(code)

    async def on_settings_updated(self):
        """
        Settings updated. Override to update anything with new settings.

        IMPORTANT: be sure to call super() as this will subscribe to listen-devices.
        """
        try:
            # Purge any existing listen device consumers
            for cons_name in self.listen_device_consumers:
                try:
                    await self.inu.js.consumer.delete(const.Streams.COMMAND, cons_name)
                except NotFoundError:
                    pass

            # Create a new subject listener (subjects may have changed with settings)
            async def on_subject_trigger(msg: model.Message):
                await self.inu.js.msg.ack(msg)

                try:
                    trg = Trigger(msg.get_payload())
                    code = int(trg.code)
                except Exception as trg_ex:
                    await self.inu.log(f"Malformed trigger payload: {type(trg_ex).__name__}: {trg_ex}", LogLevel.ERROR)
                    return

                self.logger.info(f"Trigger from {msg.subject}: code {code}")
                await self.parse_trigger_code(code, msg)

            # Even if we don't have listen subjects, listen to your own "central" address
            if not hasattr(self.inu.settings, 'listen_subjects'):
                subjects = [self.inu.get_central_id()]
            else:
                subjects = self.inu.settings.listen_subjects.split(" ")
                subjects.append(self.inu.get_central_id())

            # Create consumers for all subjects
            for subject in subjects:
                if not subject.strip():
                    continue

                try:
                    cons = await self.inu.js.consumer.create(
                        consumer.Consumer(
                            const.Streams.COMMAND,
                            consumer_cfg=consumer.ConsumerConfig(
                                filter_subject=const.Subjects.fqs(
                                    [const.Subjects.COMMAND, const.Subjects.COMMAND_TRIGGER],
                                    subject
                                ),
                                deliver_policy=consumer.ConsumerConfig.DeliverPolicy.NEW,
                                ack_wait=Time.sec_to_nano(1),
                            )
                        ), push_callback=on_subject_trigger,
                    )
                    self.listen_device_consumers.append(cons.name)
                except ErrorResponseException as e:
                    err = e.err_response
                    await self.inu.log(
                        f"Unable to subscribe to device '{subject}': [{err.code}]: {err.description}",
                        LogLevel.ERROR
                    )

            # Consumer for OTA updates
            cons = await self.inu.js.consumer.create(
                consumer.Consumer(
                    const.Streams.COMMAND,
                    consumer_cfg=consumer.ConsumerConfig(
                        filter_subject=const.Subjects.fqs(
                            [const.Subjects.COMMAND, const.Subjects.COMMAND_OTA],
                            self.inu.get_central_id()
                        ),
                        deliver_policy=consumer.ConsumerConfig.DeliverPolicy.NEW,
                        ack_wait=Time.sec_to_nano(1),
                    )
                ), push_callback=self.on_ota,
            )
            self.listen_device_consumers.append(cons.name)

            # Consumer for reboot requests
            cons = await self.inu.js.consumer.create(
                consumer.Consumer(
                    const.Streams.COMMAND,
                    consumer_cfg=consumer.ConsumerConfig(
                        filter_subject=const.Subjects.fqs(
                            [const.Subjects.COMMAND, const.Subjects.COMMAND_REBOOT],
                            self.inu.get_central_id()
                        ),
                        deliver_policy=consumer.ConsumerConfig.DeliverPolicy.NEW,
                        ack_wait=Time.sec_to_nano(1),
                    )
                ), push_callback=self.on_reboot,
            )
            self.listen_device_consumers.append(cons.name)

            await self.inu.log(f"Resubscribe to listen devices completed", LogLevel.INFO)

        except Exception as e:
            await self.inu.log(f"Error updating settings: {type(e)}: {e}", LogLevel.FATAL)

    async def on_trigger(self, code: int):
        """
        Called when a listen-device publishes a `trigger` message.

        Special codes (such as enabled, interrupt) are filtered out.
        """
        pass

    async def on_interrupt(self):
        """
        A listen-device has published an interrupt (100) code.
        """
        pass

    async def on_wait(self):
        """
        A listen-device has published an wait (103) code.
        """
        pass

    async def on_break(self):
        """
        A listen-device has published a break (104) code.
        """
        pass

    async def on_ota(self, msg: model.Message):
        """
        Received an OTA message on our central address.
        """
        try:
            ota = Ota(msg.get_payload())
        except Exception as e:
            await self.inu.js.msg.term(msg)
            await self.inu.log(f"Malformed OTA payload: {type(e).__name__}: {e}", LogLevel.ERROR)
            return

        await self.inu.js.msg.ack(msg)

        updater = OtaUpdater(self)
        await updater.perform_update(int(ota.version))

    async def on_reboot(self, msg: model.Message):
        """
        Received a reboot request.
        """
        try:
            Reboot(msg.get_payload())
        except Exception as e:
            await self.inu.js.msg.term(msg)
            await self.inu.log(f"Malformed reboot payload: {type(e).__name__}: {e}", LogLevel.ERROR)
            return

        await self.inu.js.msg.ack(msg)
        self.allow_app_tick = False

        await self.inu.log(f"Performing reboot by external request", LogLevel.WARNING)
        await self.inu.status(status="Rebooting")
        await asyncio.sleep(0.5)
        machine.reset()

    async def on_enabled_changed(self, enabled: bool):
        """
        Device enabled status was changed by a listen-device.
        """
        pass

    async def on_lock_changed(self, enabled: bool):
        """
        Device locked status was changed by a listen-device.
        """
        pass

    async def on_disconnect(self):
        """
        Disconnected with NATS server.

        Clear up consumer cache.
        """
        self.listen_device_consumers = []

    async def set_state_from_last(self, enabled_default: bool = True):
        """
        Sets the device state from the last known state.

        If there is no last known state, use the `default` value. Intended for device start-up.
        """
        try:
            last_status = await self.inu.js.msg.get_last(
                const.Streams.STATUS,
                const.Subjects.fqs(const.Subjects.STATUS, self.inu.device_id)

            )

            stat = Status(last_status.get_payload())
            await self.inu.status(enabled=stat.enabled, active=False, status="", locked=stat.locked)

        except NotFoundError:
            await self.inu.status(enabled=enabled_default, active=False, status="", locked=False)

    @staticmethod
    def map_value(x, in_min, in_max, out_min, out_max):
        return max(min((x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min, 100), 0)



import asyncio
from argparse import Namespace


class Utility:
    def __init__(self, args: Namespace):
        self.args = args
        self.exit_code = 0

    async def run(self):
        pass

    async def run_safe(self):
        try:
            await self.run()
        except asyncio.CancelledError:
            pass



from . import Settings, CooldownDevice


class TriggerDevice(Settings, CooldownDevice):
    """
    A device that sends a trigger command.
    """
    trigger_code: int = 0
    trigger_code_hint: str = "`code` to include in the trigger; may correlate to a sequence or special `code`"


class MotionSensor(TriggerDevice):
    """
    ## A motion sensor.
    """
    sensitivity: int = 50
    sensitivity_hint: str = "Sensor trip sensitivity, between 1 (low) and 100 (high)"
    sensitivity_min: int = 1
    sensitivity_max: int = 100


class RangeTrigger(TriggerDevice):
    """
    ## A range-based trip sensor.

    The device will constantly monitor the range of its sensor. If the range drops below `max_distance` the device
    will consider itself "tripped".

    Increase `wait_delay` to reduce sensitivity on false-positives.
    """
    max_distance: int = 1000
    max_distance_hint: str = "If the range drops below this value (in mm), the sensor will trigger"
    max_distance_min: int = 30

    min_distance: int = 0
    min_distance_hint: str = "Range must be above this value (and below max) for a trigger"
    min_distance_min: int = 0

    wait_delay: int = 0
    wait_delay_hint: str = "Time in ms the range must be under the max_distance value before triggering"
    wait_delay_min: int = 0


class MultiSwitch(Settings):
    """
    ## A device with one or more physical switches.

    Override code of -1 uses the default code.
    """
    trigger_code: int = 0
    trigger_code_hint: str = "Default `code` to trigger with, if not set by per-switch overrides"

    refire_delay: int = 0
    refire_delay_hint: str = "Time in ms to send another trigger when sustained (0 to disable)"

    sw_0: int = -1
    sw_0_hint: str = "Switch 0 override code"

    sw_1: int = -1
    sw_1_hint: str = "Switch 1 override code"

    sw_2: int = -1
    sw_2_hint: str = "Switch 2 override code"

    sw_3: int = -1
    sw_3_hint: str = "Switch 3 override code"

    sw_4: int = -1
    sw_4_hint: str = "Switch 4 override code"

    sw_5: int = -1
    sw_5_hint: str = "Switch 5 override code"


class Capacitor(Settings):
    """
    ## A capacitive sensor.
    """
    trigger_code: int = 0
    trigger_code_hint: str = "Default `code` to trigger with, if not set by per-switch overrides"

    sensor_low: int = 25000
    sensor_low_hint: str = "Raw low value for 0% threshold"

    sensor_high: int = 50000
    sensor_high_hint: str = "Raw high value for 100% threshold"

    trigger_on: int = 50
    trigger_on_hint: str = "Value as a percentage, to be above or below to activate trigger"

    trigger_off: int = 60
    trigger_off_hint: str = "Value as a percentage, to be above or below to deactivate trigger"

    refire_delay: int = 0
    refire_delay_hint: str = "Time in ms to send another trigger when sustained (0 to disable)"

    delay_wait: int = 0
    delay_wait_hint: str = "Time in ms to wait before triggering at threshold"


class LightSensor(Settings):
    """
    ## An ambient light sensor.
    """
    trigger_code: int = 0
    trigger_code_hint: str = "Default `code` to trigger with, if not set by per-switch overrides"

    trigger_on: int = 5
    trigger_on_hint: str = "Value in lux, to be above or below to activate trigger"

    trigger_off: int = 15
    trigger_off_hint: str = "Value in lux, to be above or below to deactivate trigger"

    refire_delay: int = 0
    refire_delay_hint: str = "Time in ms to send another trigger when sustained (0 to disable)"

    delay_wait: int = 0
    delay_wait_hint: str = "Time in ms to wait before triggering at threshold"



from . import Schema
from ..error import Malformed


class Status(Schema):
    enabled: bool = None
    active: bool = None
    locked: bool = None
    status: str = None

    def _validate(self):
        if self.enabled is None:
            raise Malformed("'enabled' cannot be None")

        if self.active is None:
            raise Malformed("'active' cannot be None")

    def __repr__(self):
        return f"enabled={self.enabled} locked={self.locked} active={self.active} status=\"{self.status}\""

    def can_act(self, allow_active=False):
        """
        Check if the state is suitable for a device to act (eg. on a trigger)

        Returns True if the device is enabled, not locked and not active.
        """
        if not allow_active and self.active:
            return False

        return self.enabled and not self.locked



from setuptools import setup

setup(
    name='inu_net',
    version="1.43.1",
    description='The Inu IoT framework',
    url='https://github.com/jordonsc/inu-py',
    author='Jordon Scott',
    author_email='jordonsc@gmail.com',
    license='MIT',
    packages=[
        'inu_net',
        'inu_net.hardware', 'inu_net.hardware.light', 'inu_net.hardware.robotics', 'inu_net.hardware.robotics.control',
        'inu_net.lib',
        'inu_net.schema', 'inu_net.schema.settings',
        'inu_net.updater',
        'inu_net.util',
    ],
    install_requires=['micro_nats'],
    classifiers=[],
)



import asyncio
import logging
import time

from ... import error, Inu, const
from ...const import LogLevel

from .control import Control
from .control.common import Select, Wait, Trigger
from .control.actuator import Move
from .control.lights import Colour, Fx

CONTROL_MAP = {
    """
    Mapping from string codes into control classes.
    """

    "S": Select,
    "SEL": Select,
    "SELECT": Select,
    "W": Wait,
    "WAIT": Wait,
    "M": Move,
    "MV": Move,
    "MOVE": Move,
    "TRG": Trigger,
    "TRIGGER": Trigger,
    "COL": Colour,
    "COLOUR": Colour,
    "COLOR": Colour,
    "FX": Fx,
}


class RoboticsDevice:
    """
    Base class for a physical device controller. Must be able to execute Control actions.
    """

    def __init__(self, inu=None, log_path="inu.robotics"):
        self.interrupted = False
        self.int_wait = False
        self.int_break = False
        self.inu = inu
        self.logger = logging.getLogger(log_path)

    async def execute(self, ctrl: Control, reverse: bool = False):
        """
        Run a control code. Non-tangible controls like SEL and WAIT will not be sent to a RoboticsDevice.

        `reverse` will undo a Control operation, while also ignoring further interrupts.
        """
        self.interrupted = False

    def set_power(self, powered: bool):
        """
        Powers or un-powers the device. May do nothing depending on if the device has a passive power state.
        """
        pass

    def req_int(self):
        """
        Inform an active operation that it has been interrupted, expecting it to reverse action already taken and abort.
        """
        self.interrupted = True

    def req_wait(self):
        """
        Inform the current/next WAIT command to reset the timer.
        """
        self.int_wait = True

    def req_break(self):
        """
        Inform the current/next WAIT command to immediately break out of the delay.
        """
        self.int_break = True

    def select_component(self, component_id):
        """
        If supported, select a sub-component of the device. This is selected by using the Control code "SEL XX:YY",
        where YY could be the `component_id`.

        `None` will be used if the control selection does not include a component (eg. "SEL XX").
        """
        print("Select component", component_id)

    async def net_log(self, msg, level=LogLevel.INFO):
        if self.inu:
            await self.inu.log(msg, level)
        else:
            if level == LogLevel.DEBUG:
                self.logger.debug(msg)
            if level == LogLevel.INFO:
                self.logger.info(msg)
            elif level == LogLevel.WARNING:
                self.logger.warning(msg)
            elif level == LogLevel.ERROR:
                self.logger.error(msg)
            elif level == LogLevel.FATAL:
                self.logger.fatal(msg)


class Robotics:
    """
    Robotics manager service.
    """

    def __init__(self, inu: Inu, power_up_delay=2500):
        self.inu = inu
        self.devices = {}
        self.logger = logging.getLogger("inu.robotics")

        # Master power state
        self.powered = False
        self.power_up_delay = power_up_delay
        self.idle_time = time.time()

        # Currently selected device (eg "A0")
        self.active_device = None

        # If the current operation has a request to interrupt, wait or break
        self.interrupted = False
        self.int_wait = False
        self.int_break = False

        # If the current operation _allows_ interruption
        self.allow_interrupt = False

    def add_device(self, device_id: str, controller: RoboticsDevice):
        """
        Add a RoboticsDevice controller to the list of actionable devices.
        """
        self.logger.info(f"Adding controller: {controller}")
        self.devices[device_id] = controller

    def select_device(self, device: Select):
        """
        Execute a SEL control code.
        """
        if device.get_device() not in self.devices:
            raise error.BadRequest(f"Device '{device.get_device()}' not registered")

        self.active_device = device.get_device()
        self.devices[self.active_device].select_component(device.get_component())

    def set_power(self, powered: bool):
        """
        Modifies the power state for all devices.
        """
        self.powered = powered

        for device in self.devices.values():
            device.set_power(powered)

    async def run(self, ctrl_str: str):
        """
        Run a control code string.
        """
        self.reset_state()
        await self.run_list(Robotics.control_array_from_string(ctrl_str))
        self.reset_state()

    async def run_list(self, control_list: list):
        """
        Run a list of operations.
        """
        int_chain = []
        last_sel = None

        # Brings device power online if it was not already
        await self.ready_devices()

        for ctrl in control_list:
            await self.inu.log(f"EXEC: {ctrl}", LogLevel.DEBUG)
            await asyncio.sleep(0)

            if ctrl.allow_interrupt():
                int_chain.append(ctrl)
                self.allow_interrupt = True
            else:
                int_chain.clear()
                self.allow_interrupt = False

                if last_sel:
                    # Important: we need to remember the last select for reversing
                    int_chain.append(last_sel)

            # Common controls
            if isinstance(ctrl, Select):
                # Select device
                self.select_device(ctrl)
                last_sel = ctrl
            elif isinstance(ctrl, Wait):
                # Wait for a given time
                start_time = time.time_ns()
                while time.time_ns() - start_time < ctrl.get_time() * 1_000_000:
                    if self.interrupted:
                        # Interrupted, drop out so the INT process can begin
                        break
                    if self.int_wait:
                        # WAIT-reset requested, restart timer
                        start_time = time.time_ns()
                        self.int_wait = False
                    if self.int_break:
                        # WAIT-break requested, immediately drop out of WAIT delay
                        self.int_break = False
                        break
                    await asyncio.sleep(0.1)
            elif isinstance(ctrl, Trigger):
                # Dispatch a trigger message
                await self.inu.command(const.Subjects.COMMAND_TRIGGER, {
                    'code': ctrl.get_code(),
                })
                await asyncio.sleep(0.1)
            elif ctrl is None:
                # Error
                await self.inu.log("Null control code provided", LogLevel.WARNING)
            else:
                # Tangible codes need to be sent to the active RoboticsDevice
                if self.active_device is None:
                    raise error.BadRequest("Attempted to execute control code with no selected device (missing SEL)")

                await self.devices[self.active_device].execute(ctrl)

            if self.interrupted:
                # Run the int_chain in reverse order..
                self.reset_state()
                await self.inu.log("Reversing ops..", LogLevel.DEBUG)
                await self.run_int_list(int_chain)

                # then run it again in normal order
                self.reset_state()
                await self.inu.log("Replaying interrupted ops..", LogLevel.DEBUG)
                await self.run_list(int_chain)
                await self.inu.log("INT seq completed", LogLevel.DEBUG)

    async def ready_devices(self):
        """
        If system is unpowered, then power it up and wait for `warmup_delay` ms.
        Resets idle time.
        """
        if not self.powered:
            self.set_power(True)
            await asyncio.sleep(self.power_up_delay / 1000)

    def reset_state(self):
        """
        Clears device state from a previous run.
        """
        self.active_device = None
        self.interrupted = False
        self.int_wait = False
        self.int_break = False
        self.allow_interrupt = False
        self.idle_time = time.time()

    async def run_int_list(self, control_list: list):
        """
        Runs a list of operations in reverse order and direction.
        """
        for ctrl in await self.prepare_int_list(control_list):
            await self.inu.log(f"REV EXEC: {ctrl}", LogLevel.DEBUG)
            await asyncio.sleep(0)

            # Non-tangible codes -
            if isinstance(ctrl, Select):
                self.select_device(ctrl)
            elif isinstance(ctrl, Wait):
                await asyncio.sleep(ctrl.get_time() / 1000)
            elif ctrl is None:
                pass
            else:
                # Tangible codes need to be sent to the active RoboticsDevice
                if self.active_device is None:
                    raise error.BadRequest("Missing SEL in INT list")

                await self.devices[self.active_device].execute(ctrl, reverse=True)

    async def prepare_int_list(self, control_list: list):
        """
        Reverse the list, moving SEL statements to the front of their controls.

        Skips Wait controls.
        """
        int_list = []
        buffer = []
        # Skip the last element as that would have been partially completed and already reversed
        for ctrl in reversed(control_list[:-1]):
            if isinstance(ctrl, Select):
                int_list.append(ctrl)
                int_list += buffer
                buffer = []
            elif isinstance(ctrl, Wait) or ctrl is None:
                pass
            else:
                buffer.append(ctrl)

        if len(buffer) > 0:
            await self.inu.log("INT list has no preceding SEL", LogLevel.WARNING)
            int_list += buffer

        return int_list

    def req_int(self) -> bool:
        """
        Interrupt the current operation, halting and reversing.

        Returns True if the interrupt was accepted.
        """
        if self.active_device and self.allow_interrupt:
            self.interrupted = True
            self.devices[self.active_device].req_int()
            return True
        else:
            return False

    def req_wait(self) -> bool:
        """
        Reset the current WAIT timer.

        Returns True if the wait was accepted.
        """
        if self.active_device:
            self.int_wait = True
            self.devices[self.active_device].req_wait()
            return True
        else:
            return False

    def req_break(self) -> bool:
        """
        Reset the current WAIT timer.

        Returns True if the wait was accepted.
        """
        if self.active_device:
            self.int_break = True
            self.devices[self.active_device].req_break()
            return True
        else:
            return False

    def get_idle_time(self) -> float:
        """
        Returns the time in seconds that the device has been idle.
        """
        return time.time() - self.idle_time

    @staticmethod
    def control_from_string(ctrl: str) -> Control:
        """
        Construct a Control class from a string.
        """
        code = ctrl.strip().upper().split(" ")[0]
        if code not in CONTROL_MAP:
            raise error.BadRequest(f"Unknown control code: {ctrl}")

        return CONTROL_MAP[code](ctrl)

    @staticmethod
    def control_array_from_string(ctrl_list: str) -> list:
        """
        Construct a list of Control classes from a full control string.

        Control codes are delimited by a semi-colon (Control.DELIMITER).
        """
        arr = []
        cmds = ctrl_list.split(Control.DELIMITER)

        for cmd in cmds:
            arr.append(Robotics.control_from_string(cmd))

        return arr



from . import Control
from inu import error


class Move(Control):
    """
    Move the device, such as an actuator, by a given distance at a given speed.

    MV <distance> <speed>
    """
    CONTROL_CODE = "MV"
    ALIASES = ["M", "MV", "MOVE"]

    def __init__(self, ctrl: str = None):
        super().__init__(ctrl)

        if self.code not in self.ALIASES or len(self.args) != 2:
            raise error.Malformed(f"Invalid {self.CONTROL_CODE} control: {ctrl}")

    def get_distance(self) -> int:
        """
        Distance of the move operation, in mm.
        """
        return int(self.args[0])

    def get_speed(self) -> int:
        """
        Speed to move the actuator in mm/s.
        """
        return int(self.args[1])

    def __repr__(self):
        return f"MV {self.get_distance()} mm @ {self.get_speed()} mm/s"



from inu import error


class ColourCode:
    # Colour mappings courtesy of Copilot :)
    COLOUR_TABLE = {
        "BLACK": "#000000",
        "WHITE": "#FFFFFF",
        "RED": "#FF0000",
        "GREEN": "#00FF00",
        "BLUE": "#0000FF",
        "YELLOW": "#FFFF00",
        "CYAN": "#00FFFF",
        "MAGENTA": "#FF00FF",
        "SILVER": "#C0C0C0",
        "GRAY": "#808080",
        "MAROON": "#800000",
        "OLIVE": "#808000",
        "LIME": "#00FF00",
        "TEAL": "#008080",
        "AQUA": "#00FFFF",
        "NAVY": "#000080",
        "FUCHSIA": "#FF00FF",
        "PURPLE": "#800080",
        "ORANGE": "#FFA500",
        "BROWN": "#A52A2A",
        "TURQUOISE": "#40E0D0",
        "GOLD": "#FFD700",
        "PINK": "#FFC0CB",
        "VIOLET": "#EE82EE",
        "INDIGO": "#4B0082",
        "BEIGE": "#F5F5DC",
        "KHAKI": "#F0E68C",
        "TAN": "#D2B48C",
        "HONEYDEW": "#F0FFF0",
        "AZURE": "#F0FFFF",
        "LAVENDER": "#E6E6FA",
        "CORAL": "#FF7F50",
        "SALMON": "#FA8072",
        "TOMATO": "#FF6347",
        "ORCHID": "#DA70D6",
        "PERU": "#CD853F",
        "SEASHELL": "#FFF5EE",
        "CHOCOLATE": "#D2691E",
        "LINEN": "#FAF0E6",
    }

    def __init__(self, *args):
        """
        Create a colour code object from a hex or RGB string. The colour code may contain an optional 4th parameter
        which is the brightness.

        ex:
        ColourCode("#FF0000")  # Red
        ColourCode("255,0,0")  # Red
        ColourCode("#FF0000A0")  # Red, reduced brightness
        ColourCode("255,0,0,160")  # Red, reduced brightness
        ColourCode("RED")  # Red
        ColourCode(255, 0, 0, 255)  # Red

        :param col: Can be a colour name, an HTML-style hex code, or an RGB code.
        """
        self.r = 0
        self.g = 0
        self.b = 0
        self.x = 255

        if len(args) == 4:
            self.r, self.g, self.b, self.x = [ColourCode.check_bounds(int(part)) for part in args]
        elif len(args) == 1:
            col = str(args[0]).upper().strip()

            # Convert colour names to hex values
            if col in self.COLOUR_TABLE:
                col = self.COLOUR_TABLE[col]

            if col[0] == '#':
                # Hex code, HTML style
                self.decode_hex(col[1:])
            elif "," in col:
                # RGB (decimal) code
                self.decode_rgb(col)
            else:
                raise error.Malformed(f"Invalid colour code: {col}")
        else:
            raise error.Malformed("Invalid number of arguments")

    def decode_hex(self, col: str):
        if len(col) != 6 and len(col) != 8:
            raise error.Malformed("Invalid hex color code")

        self.r, self.g, self.b = [int(col[i:i + 2], 16) for i in range(0, 6, 2)]
        self.x = int(col[6:8], 16) if len(col) == 8 else 255

    def decode_rgb(self, col: str):
        parts = col.split(",")

        if len(parts) < 3 or len(parts) > 4:
            raise error.Malformed("Invalid RGB color code")

        self.r, self.g, self.b = [ColourCode.check_bounds(int(part.strip())) for part in parts[:3]]
        self.x = ColourCode.check_bounds(int(parts[3].strip())) if len(parts) == 4 else 255

    def unpack(self):
        return self.r, self.g, self.b, self.x

    @staticmethod
    def check_bounds(value):
        if 0 <= value <= 255:
            return value
        else:
            raise error.Malformed("RGB values must be between 0 and 255 inclusive")

    def __repr__(self):
        return f"#{self.r:02x}{self.g:02x}{self.b:02x}{self.x:02x}"



class InuError(Exception):
    pass


class BadRequest(InuError):
    pass


class NoConnection(InuError):
    pass


class Malformed(InuError):
    pass


class InvalidDeviceId(InuError):
    pass


class UnsupportedDeviceType(InuError):
    pass


class DeviceAlert(InuError):
    pass


class LimitHalt(InuError):
    pass



import asyncio
import time

from machine import Pin, PWM
from ..switch import Switch
from ... import error
from ...const import LogLevel

from . import RoboticsDevice
from .control import Control
from .control.actuator import Move


class StepperDriver:
    def __init__(self, pulse, direction, enabled, alert=None):
        self.pulse = Pin(pulse, Pin.OUT)
        self.direction = Pin(direction, Pin.OUT)
        self.enabled = Pin(enabled, Pin.OUT)

        if alert is not None:
            self.alert = Switch(alert)
        else:
            self.alert = None


class Screw:
    def __init__(self, steps_per_rev: int = 1600, screw_lead: int = 8, forward: int = 1):
        # Number of steps per revolution
        self.steps_per_rev = steps_per_rev

        # Screw lead (distance actuator moves for 1 rotation)
        self.screw_lead = screw_lead

        # Driver direction that is "forward"
        self.forward = forward

    def __repr__(self):
        return f"steps/rev: {self.steps_per_rev}; lead: {self.screw_lead}"


class OpVector:
    def __init__(self, min_speed, speed, distance, ramp_accel):
        self.min_speed = min_speed
        self.speed = speed
        self.total_displacement = distance
        self.ramp_accel = max(self.get_min_ramp_accel(), ramp_accel)

        # Time spent ramping up/down
        self.ramp_time = speed / self.ramp_accel
        # Displacement during a single ramp phase
        self.ramp_displacement = ((speed - min_speed) + min_speed) / 2 * self.ramp_time
        # Displacement during the full-speed phase
        self.full_displacement = distance - (self.ramp_displacement * 2)

        self.ramp_time = self.ramp_time * 10 ** 9  # convert to NS
        self.full_spd_time = self.full_displacement / speed * 10 ** 9
        self.op_time = self.full_spd_time + (self.ramp_time * 2)

    def get_min_ramp_accel(self):
        """
        Returns the minimum acceleration we need to ramp at in order to hit full-speed during the operation.
        """
        return (2 * self.speed ** 2) / self.total_displacement

    def __repr__(self):
        return f"<op ramp_in={round(self.ramp_time * 10 ** -9, 2)} " + \
            f"full_spd={round(self.full_spd_time * 10 ** -9, 2)} " + \
            f"ramp_out={round(self.ramp_time * 10 ** -9, 2)} " + \
            f"accel={self.ramp_accel} " + \
            f"op_time={round(self.op_time * 10 ** -9, 2)}>"


class Actuator(RoboticsDevice):
    """
    Moves an actuator forward or backwards.

    Supports undoing partial operations on interrupt.
    """
    CONFIG_ALIASES = ["actuator", "stepper"]

    # Required time remaining in an operation (in nanoseconds) to allow yielding CPU
    MIN_SLEEP_TIME = 0.25 * 10 ** 9  # 0.25 seconds

    # Time to pause when interrupted before reversing
    INT_PAUSE_TIME = 0.5

    class DisplacementPhase:
        RAMP_UP = 0
        FULL_SPEED = 1
        RAMP_DOWN = 2
        END = 3
        LIMIT_HALT = 4

    def __init__(self, driver: StepperDriver, screw: Screw, ramp_speed: int = 150, halt_ramp_speed: int = 300,
                 fwd_stop: Switch = None, rev_stop: Switch = None, allow_sleep: bool = True, inu=None):
        """
        `ramp_speed` is the acceleration rate in mm/s to start/stop the stepper.
        `allow_sleep` will allow the device to yield CPU if there is more than MIN_SLEEP_TIME nanoseconds remaining in
        the operation.
        """
        super().__init__(inu=inu, log_path="inu.robotics.actuator")

        self.driver = driver
        self.screw = screw

        self.driver.pulse.off()
        self.driver.direction.off()
        self.driver.enabled.off()

        self.fwd_stop = fwd_stop
        self.rev_stop = rev_stop

        self.allow_sleep = allow_sleep
        self.ramp_accel = ramp_speed
        self.halt_accel = halt_ramp_speed

        # Displacement of last operation
        self.displacement = 0

    def on(self):
        """
        Power the stepper motor preventing rotation while idle.
        """
        self.set_power(True)

    def off(self):
        """
        Un-power the stepper motion, allowing rotation while idle.
        """
        self.set_power(False)

    def set_power(self, on: bool):
        """
        Set the 'enable' option on the stepper driver. If set to True, the motor will not allow rotation while not in
        use. If disabled, the motor will be permitted rotation while not active.

        NB: Calling `drive()` will enable the motor power, but not disable it following.
        """
        self.driver.enabled.value(on)

    def distance_to_steps(self, displacement: float) -> int:
        """
        Calculate the number of stepper motor steps for the given actuator displacement in mm.
        """
        return round((displacement / self.screw.screw_lead) * self.screw.steps_per_rev)

    def pulse_rate_from_speed(self, speed: float) -> int:
        """
        Calculate the pulses per second from a speed.
        """
        return round(speed / self.screw.screw_lead * self.screw.steps_per_rev)

    async def drive(self, distance: float, speed: float = 10, direction: int = 1, ignore_int=False):
        """
        Move the actuator by a given distance.

        distance:  distance to move the actuator in mm
        speed:     speed to move the actuator in mm/s
        direction: direction of stepper motor, 1 == "forward"

        CAUTION: do not drive immediately after a high-speed drive, or change rotation back-to-back.
                 a delay of 0.2s is recommended.

        If the driver isn't enabled, it will be enabled. Does not disable upon completion.
        """
        fwd = direction == 1

        # Flip the direction if the screw direction is reversed
        direction = int(not (self.screw.forward ^ direction))

        if self.driver.enabled.value() == 0:
            self.set_power(True)
            await asyncio.sleep(0.25)

        if self.driver.direction.value() != direction:
            self.driver.direction.value(direction)

        # Don't even start the stepper if the limiter is already triggered
        if fwd and self.fwd_stop and await self.fwd_stop.check_state(no_delay=True):
            return

        if not fwd and self.rev_stop and await self.rev_stop.check_state(no_delay=True):
            return

        # Calculate ramp times
        op = OpVector(min_speed=1, speed=speed, distance=distance, ramp_accel=self.ramp_accel)
        self.logger.info(op)
        if op.ramp_accel > self.ramp_accel:
            self.logger.warning(f"Required operation acceleration ({op.ramp_accel} is greater than configured " +
                                f"acceleration ({self.ramp_accel}")

        self.displacement = 0
        phase = self.DisplacementPhase.RAMP_UP
        run_time = 0
        current_speed = op.min_speed

        last_tick = time.time_ns()
        start_time = last_tick

        pwm = PWM(self.driver.pulse, freq=self.pulse_rate_from_speed(current_speed), duty=512)

        while True:
            if self.displacement >= distance:
                break

            # An interrupt signal has been received, we need to stop and reverse the action
            if not ignore_int and self.interrupted:
                # We'll exit cleanly and the caller can deal with the reverse op
                # NB: we can't interrupt during ramping - we can only cut short the full-speed phase
                if phase == self.DisplacementPhase.END:
                    break
                elif phase == self.DisplacementPhase.FULL_SPEED:
                    await self.net_log("Interrupted", LogLevel.DEBUG)
                    phase = self.DisplacementPhase.RAMP_DOWN
                    # Update full-speed time to adjust the ramping calcs
                    op.full_spd_time = run_time - op.ramp_time

            # Check for an alert from the controller
            if self.driver.alert is not None and await self.driver.alert.check_state():
                # This should be wrapped in a handler that will dispatch an appropriate alert and shutdown all
                # robotics functions (depending on context)
                pwm.deinit()
                raise error.DeviceAlert()

            # Check limiters
            if phase != self.DisplacementPhase.LIMIT_HALT:
                if fwd and self.fwd_stop and await self.fwd_stop.check_state():
                    await self.net_log("Forward limiter halt", LogLevel.DEBUG)
                    # phase = self.DisplacementPhase.LIMIT_HALT
                    break
                if not fwd and self.rev_stop and await self.rev_stop.check_state():
                    await self.net_log("Reverse limiter halt", LogLevel.DEBUG)
                    # phase = self.DisplacementPhase.LIMIT_HALT
                    break

            tick = time.time_ns()
            tick_time = tick - last_tick
            if tick_time == 0:
                continue
            else:
                last_tick = tick

            self.displacement += current_speed * tick_time * 10 ** -9
            run_time = tick - start_time

            # Update PWM speed/phase
            if phase == self.DisplacementPhase.RAMP_UP:
                # Accelerating to full speed
                pos = run_time / op.ramp_time
                current_speed = min(((op.speed - op.min_speed) * pos) + op.min_speed, op.speed)
                pwm.freq(self.pulse_rate_from_speed(current_speed))
                if current_speed >= op.speed:
                    self.logger.info(f"Move to full-speed phase at {run_time * 10 ** -9} s; pos {pos}")
                    phase = self.DisplacementPhase.FULL_SPEED
            elif phase == self.DisplacementPhase.FULL_SPEED:
                # Main run phase at intended speed
                if run_time >= op.ramp_time + op.full_spd_time:
                    self.logger.info(f"Move to ramp-down phase at {run_time * 10 ** -9} s")
                    phase = self.DisplacementPhase.RAMP_DOWN
            elif phase == self.DisplacementPhase.RAMP_DOWN:
                # Decelerating to come to a halt
                pos = 1 - ((run_time - op.full_spd_time - op.ramp_time) / op.ramp_time)
                current_speed = max(((op.speed - op.min_speed) * pos) + op.min_speed, op.min_speed)
                pwm.freq(self.pulse_rate_from_speed(current_speed))
                if current_speed <= op.min_speed:
                    self.logger.info(f"Move to end phase at {run_time * 10 ** -9} s; pos {pos}")
                    phase = self.DisplacementPhase.END

                    # if we've been interrupted, exit immediately - don't attempt to finish the full distance
                    if not ignore_int and self.interrupted:
                        break
            elif phase == self.DisplacementPhase.LIMIT_HALT:
                # A limiter has been triggered, halt as quickly as we possibly can
                current_speed -= self.halt_accel * (tick_time * 10 ** -9)

                if current_speed <= op.min_speed:
                    break
                else:
                    pwm.freq(self.pulse_rate_from_speed(current_speed))

            elif phase == self.DisplacementPhase.END:
                # If for some reason we're still under the full distance (shouldn't happen), just crawl the rest of
                # the way at op.min_speed (we're assuming we're nanometers away..)
                pass

            # Ramp-down/end is time-sensitive, do not allow sleeping (passing CPU to other tasks)
            if phase < self.DisplacementPhase.RAMP_DOWN:
                await asyncio.sleep(0)

        pwm.deinit()
        await self.net_log(f"Done in {run_time * 10 ** -9} s; displacement: {self.displacement}", LogLevel.DEBUG)

    async def execute(self, ctrl: Control, reverse: bool = False):
        await super().execute(ctrl)

        if isinstance(ctrl, Move):
            if reverse:
                direction = int(ctrl.get_distance() < 0)
            else:
                direction = int(ctrl.get_distance() >= 0)

            distance = abs(ctrl.get_distance())
            await self.drive(distance, ctrl.get_speed(), direction, ignore_int=reverse)

            if not reverse and self.interrupted:
                await asyncio.sleep(self.INT_PAUSE_TIME)
                self.logger.info(f"Interrupt reverse for {self.displacement} mm")
                await self.drive(self.displacement, ctrl.get_speed(), int(not bool(direction)), ignore_int=True)

    def __repr__(self):
        return f"Stepper <{self.screw}>"
